"""
NRV-materials
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import os

from ..backend.log_interface import rise_warning
from ..backend.NRV_Class import NRV_class

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

# get the built-in material librairy
dir_path = os.environ["NRVPATH"] + "/_misc"
material_library = os.listdir(dir_path + "/materials/")

###############
## Functions ##
###############


def is_mat(mat):
    """
    check if an object is a material, return True if yes, else False

    Parameters
    ----------
    mat : object
        object to test

    Returns
    -------
    bool
        True it the type is a material object
    """
    return isinstance(mat, material)


def get_mat_file_as_dict(fname):
    """
    Open .mat material librairy file and return all lines as a dictionnary

    Returns
    -------
    d   : dictionnary
        physical properties of the material
    """
    d = {}
    with open(fname, "r") as f:
        for line in f:
            (key, value) = line.split()
            d[key] = value
    return d


def load_material(f_material):
    """
    Load a material by its name. if the name is in the NRV2 material librairy, the extention is
    automatically added.

    Parameters
    ----------
    f_material  : str
        either material name if the material is in the NRV2 Librairy or path to the corresponding\
        .mat material file
    """
    # load from librairy or from file
    f_in_librairy = str(f_material) + ".mat"
    if f_in_librairy in material_library:
        mat_file = get_mat_file_as_dict(dir_path + "/materials/" + f_in_librairy)
    else:
        mat_file = get_mat_file_as_dict(f_material)
    # creat material instance
    mat_obj = material()
    if "name" in mat_file:
        mat_obj.set_name(mat_file["name"])
    if "source" in mat_file:
        mat_obj.set_source(mat_file["source"])
    if "sigma_xx" in mat_file:
        mat_obj.set_anisotropic_conductivity(
            mat_file["sigma_xx"], mat_file["sigma_yy"], mat_file["sigma_zz"]
        )
    elif "sigma" in mat_file:
        mat_obj.set_isotropic_conductivity(mat_file["sigma"])
    else:
        rise_warning(
            "loading a material with 0 conductivity, \
            this may induce further division by 0"
        )
    return mat_obj


####################
## material class ##
####################
class material(NRV_class):
    """
    a class for material, where all the physical properties constants are stored.
    """

    def __init__(self):
        """
        material instantiation
        """
        super().__init__()
        self.name = ""
        self.source = ""
        self.isotrop_cond = True
        self.sigma = 0
        self.sigma_xx = 0
        self.sigma_yy = 0
        self.sigma_zz = 0

    def save_material(self, save=False, fname="material.json"):
        rise_warning("save_material is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_material(self, data="material.json"):
        rise_warning("load_material is a deprecated method use load")
        self.load(data=data)

    def set_name(self, name):
        """
        set a name to a material

        Parameters
        ----------
        name :  str
            name of the material
        """
        self.name = name

    def set_source(self, source):
        """
        set a source to a material. Note that this source is a string without spaces

        Parameters
        ----------
        source  : str
            scientific reference to the value, for clarity only
        """
        self.source = source

    def set_isotropic_conductivity(self, sigma):
        """
        set the conductivity for an isotropic material

        Parameters
        ----------
        sigma   : float
            conductivity in S/m
        """
        self.isotrop_cond = True
        self.sigma = float(sigma)

    def set_anisotropic_conductivity(self, sigma_xx, sigma_yy, sigma_zz):
        """
        set the conductivity tensor for an anisotropic material

        Parameters
        ----------
        sigma_xx    : float
            conductivity along the x axis, in S/m
        sigma_yy    : float
            conductivity along the y axis, in S/m
        sigma_zz    : float
            conductivity along the z axis, in S/m
        """
        self.isotrop_cond = False
        self.sigma_xx = float(sigma_xx)
        self.sigma_yy = float(sigma_yy)
        self.sigma_zz = float(sigma_zz)

    def is_isotropic(self):
        """
        check that the material is isotropic or not, return true if isotropic, else false

        Returns
        -------
        bool    :
            True if the per
        """
        return self.isotrop_cond



"""
NRV-extracellular contexts
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler

import numpy as np

from ..backend.log_interface import rise_error, rise_warning
from ..backend.MCore import MCH, synchronize_processes
from ..backend.NRV_Class import NRV_class
from ..utils.units import cm, m
from .materials import load_material

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

MRG_fiberD = np.asarray([1, 2, 5.7, 7.3, 8.7, 10.0, 11.5, 12.8, 14.0, 15.0, 16.0])
MRG_nodeD = np.asarray([0.7, 1.4, 1.9, 2.4, 2.8, 3.3, 3.7, 4.2, 4.7, 5.0, 5.5])


def is_recording_point(point):
    """
    Check if the specified object is a recording point
    """
    return isinstance(point, recording_point)


def is_recorder(rec):
    """
    Check if the specified object is a recorder
    """
    return isinstance(rec, recorder)


def NodeD_interpol(diameter):
    """
    Compute the MRG Node diameters

    Parameters
    ----------
    diameter    : float
        diameter of the unmylinated axon to implement, in um
    Returns
    -------
    nodeD       : float
    """
    if diameter in MRG_fiberD:
        index = np.where(MRG_fiberD == diameter)[0]
        nodeD = MRG_nodeD[index]
    else:
        # create fiting polynomyals
        nodeD_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_nodeD, 3))
        nodeD = nodeD_poly(diameter)
    return float(nodeD)


class recording_point(NRV_class):
    """
    Object equivalent to a point source electrode for extracellular potential recording only (No stimulation)
    """

    def __init__(self, x=0, y=0, z=0, ID=0, method="PSA"):
        """
        Instantiation of a recording point.

        Parameters
        ----------
        x       : float
            x position of the recording point, in um
        y       : float
            y position of the recording point, in um
        z       : float
            z position of the recording point, in um
        ID      : int
            electrode identification number, set to 0 by default
        method  : string
            electrical potential approximation method, can be "PSA" (Point Source Approximation)
            or "LSA" (Line Source Approximation).
            set to "PSA" by default. Note that if LSA is requested with an anisotropic material, computation
            will automatically be performed using "PSA"
        """
        super().__init__()
        self.type = "recording_point"
        # properties
        self.ID = ID
        self.x = x
        self.y = y
        self.z = z
        self.method = method
        # footprints
        self.footprints = (
            {}
        )  # footprints are stored for each axon with the key beeing the axon's ID
        self.init = False
        self.recording = None

    def save_recording_point(self, save=False, fname="recording_point.json"):
        rise_warning("save_recording_point is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_recording_point(self, data="recording_point.json"):
        rise_warning("load_recording_point is a deprecated method use load")
        self.load(data=data)

    def translate(self, x=None, y=None, z=None):
        """
        Move recording point by translation

        Parameters
        ----------
        x   : float
            x axis value for the translation in um
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        if x is not None:
            self.x += x
        if y is not None:
            self.y += y
        if z is not None:
            self.z += z

    def get_ID(self):
        """
        get the ID of a recording point

        Returns
        -------
        id : int
            ID number of the recording point
        """
        return self.ID

    def get_method(self):
        """
        get the method for a recording point

        Returns
        -------
        method : str
            name of the method, either "PSA" for point source approximation or "LSA" for line Source Approximation
        """
        return self.method

    def set_method(self, method):
        """
        Set the approximation to compute the extracellular potential

        Parameters
        ----------
        method : str
            name of the method, either "PSA" for point source approximation or "LSA" for line Source Approximation
        """
        self.method = method

    def compute_PSA_isotropic_footprint(
        self,
        x_axon,
        y_axon,
        z_axon,
        d,
        ID,
        sigma,
        myelinated=False,
    ):
        """
        Compute the footprint for Point Source approximation an isotropic material

        Parameters
        ----------
        x_axon  : array
            array of the positions of the axon along the x-axis at which there is a current source, in [um]
        y_axon  : float
            y-axis position of the axon, in [um]
        z-axon  : float
            z-axis position of the axon, in [um]
        d       : float
            diameter of the axon, in [um]
        ID      : int
            axon ID number
        sigma   : float
            conductivity of the isotropic extracellular material, in [S.m]
        """
        if myelinated:
            d = NodeD_interpol(d)
            surface = np.pi * (d / cm) * (1 / cm)
        else:
            surface = np.pi * (d / cm) * (np.gradient(x_axon) / cm)
        electrical_distance = (
            4
            * np.pi
            * sigma
            * (
                ((self.x - x_axon) / m) ** 2
                + ((self.y - y_axon) / m) ** 2
                + ((self.z - z_axon) / m) ** 2
            )
            ** 0.5
        )
        self.footprints[str(ID)] = np.divide(surface, electrical_distance)

    def compute_PSA_anisotropic_footprint(
        self,
        x_axon,
        y_axon,
        z_axon,
        d,
        ID,
        sigma_xx,
        sigma_yy,
        sigma_zz,
        myelinated=False,
    ):
        """
        Compute the footprint for Point Source approximation an anisotropic material

        Parameters
        ----------
        x_axon      : array
            array of the positions of the axon along the x-axis at which there is a current source, in [um]
        y_axon      : float
            y-axis position of the axon, in [um]
        z-axon      : float
            z-axis position of the axon, in [um]
        d           : float
            diameter of the axon, in [um]
        ID          : int
            axon ID number
        sigma_xx    : float
            conductivity of the isotropic extracellular material alog the x-axis, in [S.m]
        sigma_yy    : float
            conductivity of the isotropic extracellular material alog the y-axis, in [S.m]
        sigma_zz    : float
            conductivity of the isotropic extracellular material alog the z-axis, in [S.m]
        """
        if myelinated:
            d = NodeD_interpol(d)
            surface = np.pi * (d / cm) * (1 / cm)
            d_internode = np.gradient(x_axon)
        else:
            surface = surface = np.pi * (d / cm) * (np.gradient(x_axon) / cm)
            d_internode = np.gradient(x_axon)

        sx = sigma_yy * sigma_zz
        sy = sigma_xx * sigma_zz
        sz = sigma_xx * sigma_yy

        electrical_distance = (
            4
            * np.pi
            * (
                (sx * (self.x - x_axon) / m) ** 2
                + (sy * (self.y - y_axon) / m) ** 2
                + (sz * (self.z - z_axon) / m) ** 2
            )
            ** 0.5
        )
        self.footprints[str(ID)] = np.divide(surface, electrical_distance)

    def compute_LSA_isotropic_footprint(
        self,
        x_axon,
        y_axon,
        z_axon,
        d,
        ID,
        sigma,
        myelinated=False,
    ):
        """
        Compute the footprint for Linear Source approximation an isotropic material

        Parameters
        ----------
        x_axon  : array
            array of the positions of the axon along the x-axis at which there is a current source, in [um]
        y_axon  : float
            y-axis position of the axon, in [um]
        z-axon  : float
            z-axis position of the axon, in [um]
        d       : float
            diameter of the axon, in [um]
        ID      : int
            axon ID number
        sigma   : float
            conductivity of the isotropic extracellular material, in [S.m]
        """
        if myelinated:
            d = NodeD_interpol(d)
            surface = np.pi * (d / cm) * (1 / cm)
            d_internode = np.gradient(x_axon)
        else:
            surface = surface = np.pi * (d / cm) * (np.gradient(x_axon) / cm)
            d_internode = np.gradient(x_axon)

        dist = (((self.x - x_axon) / m) ** 2 + ((self.y - y_axon) / m) ** 2) ** 0.5
        electrical_distance = (4 * np.pi * (d_internode / m) * sigma) / np.log(
            abs(
                (dist - (self.x - x_axon) / m)
                / (
                    (
                        (
                            (d_internode + (self.x - x_axon)) ** 2
                            + (self.y - y_axon) ** 2
                        )
                        ** 0.5
                    )
                    / m
                    - (d_internode + (self.x - x_axon)) / m
                )
            )
        )
        self.footprints[str(ID)] = np.divide(surface, electrical_distance)

    def init_recording(self, N_points):
        """
        Initializes the recorded extracellular potential. if a recording already exists,
        nothing is performed (usefull at the multi-axons level for instance)

        Parameters
        ----------
        N_points : int
            length of the extracellular potential vector along temporal dimension
        """
        if not self.init:
            self.recording = np.zeros(N_points)
            self.init = True
        else:
            if len(self.recording) != N_points:
                rise_error(
                    "Trying to compute an extracellular potential of a wrong temporal size"
                )

    def reset_recording(self, N_points):
        """
        Sets the recorded extracellular potential to zero whatever the conditions

        Parameters
        ----------
        N_points : int
            length of the extracellular potential vector along temporal dimension
        """
        self.recording = np.zeros(N_points)

    def add_axon_contribution(self, I_membrane, ID):
        """
        Adds the and axon extracellular potential to the recording computed as the matrix product of
        I_membrane computed from neuron with the footprint computed below and stored in the object
        for a specific axons ID.

        Parameters
        ----------
        I_membrane  : array
            membrane current as a matrix over times (columns) and position (lines)
            in mA/cm2
        ID          : int
            axon ID as footprint are stored in a dictionnary with their ID as key
        """
        self.recording += np.matmul(np.transpose(I_membrane), self.footprints[str(ID)])


class recorder(NRV_class):
    """
    Object for recording extracellular potential of axons.
    """

    def __init__(self, material=None):
        """
        Instantiation of an extracellular potential recording mechanism. A mecanism can store recording points,
        be associated with a material and properties. The mechanism will perform the extracellular potential
        computation at each point for an axon when a simulation is performed.
        """
        super().__init__()
        self.type = "recorder"
        self.material = None
        self.is_isotropic = True
        self.t = None
        self.t_init = False
        # if no material specified, sigma is 1S/m, else everything is loaded from signal
        if material == None:
            self.sigma = 1
            self.isotropic = True
            self.material = None
            self.sigma_xx = None
            self.sigma_yy = None
            self.sigma_zz = None
        else:
            self.material = material
            temporary_material = load_material(self.material)
            if temporary_material.is_isotropic():
                self.isotropic = True
                self.sigma = temporary_material.sigma
                self.sigma_xx = None
                self.sigma_yy = None
                self.sigma_zz = None
            else:
                self.isotropic = False
                self.sigma = None
                self.sigma_xx = temporary_material.sigma_xx
                self.sigma_yy = temporary_material.sigma_yy
                self.sigma_zz = temporary_material.sigma_zz
        # for internal use
        self.recording_points = []

    def save_recorder(self, save=False, fname="recorder.json"):
        rise_warning("save_recorder is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_recorder(self, data="recorder.json"):
        rise_warning("load_recorder is a deprecated method use load")
        self.load(data=data)

    def is_empty(self):
        """
        check if a recorder has no recording points (empty) or not.

        Returns
        -------
        """
        return self.recording_points == []

    def translate(self, x=None, y=None, z=None):
        """
        Move recorder rec points by group translation

        Parameters
        ----------
        x   : float
            x axis value for the translation in um
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        for rec_p in self.recording_points:
            rec_p.translate(x=x, y=y, z=z)

    def set_time(self, t_vector):
        """
        set the time vector of a recorder

        t_vector : array
            array of recording times, in ms
        """
        if not self.t_init:
            self.t = t_vector
            self.t_init = True

    def add_recording_point(self, point):
        """
        add an object of type recording_point to the list of recording points.

        Parameters
        ----------
        point : recording_point
            recording point to add to the recording points list
        """
        if is_recording_point(point):
            self.recording_points.append(point)

    def set_recording_point(self, x, y, z, method="PSA"):
        """
        Set a recording point at a given location and add to the recording points list

        Parameters
        ----------
        x       : float
            x position of the recording point, in um
        y       : float
            y position of the recording point, in um
        z       : float
            z position of the recording point, in um
        Parameters
        method  : string
            electrical potential approximation method, can be "PSA" (Point Source Approximation)
            or "LSA" (Line Source Approximation).
            set to "PSA" by default. Note that if LSA is requested with an anisotropic material,
            computation
            will automatically be performed using "PSA"
        """
        if self.is_empty():
            new_point = recording_point(x, y, z, method=method)
        else:
            lowest_ID = self.recording_points[-1].get_ID()
            new_point = recording_point(x, y, z, ID=lowest_ID + 1, method=method)
        self.add_recording_point(new_point)

    def set_recording_zplane(
        self, x_min, x_max, y_min, y_max, z, dx=10, dy=10, method="PSA"
    ):
        """
        Generate equaly spaced recording points in the z plane

        Parameters
        ----------
        x_min   : float
            minimal x postion for recording points, in um
        x_max   : float
            maximal x postion for recording points, in um
        y_min   : float
            minimal y postion for recording points, in um
        y_max   : float
            maximal y postion for recording points, in um
        z       : float
            fixed z position for recording points
        dx      : float
            distance between recording points on the x coordinate, in um
        dy      : float
            distance between recording points on the y coordinate, in um
        method  : string
            electrical potential approximation method, can be "PSA" (Point Source Approximation)
            or "LSA" (Line Source Approximation).
            set to "PSA" by default. Note that if LSA is requested with an anisotropic material,
            computation
            will automatically be performed using "PSA"
        """
        if np.abs(x_max - x_min) < dx:
            dx = np.abs(x_max - x_min) / 10
            rise_warning("dx too large, changed automatically to " + str(dx) + " um")
        if np.abs(y_max - y_min) < dy:
            dy = np.abs(y_max - y_min) / 10
            rise_warning("dy too large, changed automatically to " + str(dy) + " um")
        x_positions = np.arange(x_min, x_max, dx)
        y_positions = np.arange(y_min, y_max, dy)
        for x in x_positions:
            for y in y_positions:
                self.set_recording_point(x, y, z, method=method)

    def set_recording_yplane(
        self, x_min, x_max, y, z_min, z_max, dx=10, dz=10, method="PSA"
    ):
        """
        Generate equaly spaced recording points in the y plane

        Parameters
        ----------
        x_min   : float
            minimal x postion for recording points, in um
        x_max   : float
            maximal x postion for recording points, in um
        y       : float
            fixed y position for recording points
        z_min   : float
            minimal z postion for recording points, in um
        z_max   : float
            maximal z postion for recording points, in um
        dx      : float
            distance between recording points on the x coordinate, in um
        dz      : float
            distance between recording points on the z coordinate, in um
        method  : string
            electrical potential approximation method, can be "PSA" (Point Source Approximation)
            or "LSA" (Line Source Approximation). set to "PSA" by default.
            Note that if LSA is requested with an anisotropic material,computation
            will automatically be performed using "PSA"
        """
        if np.abs(x_max - x_min) < dx:
            dx = np.abs(x_max - x_min) / 10
            rise_warning("dx too large, changed automatically to " + str(dx) + " um")
        if np.abs(z_max - z_min) < dz:
            dz = np.abs(z_max - z_min) / 10
            rise_warning("dz too large, changed automatically to " + str(dz) + " um")
        x_positions = np.arange(x_min, x_max, dx)
        z_positions = np.arange(z_min, z_max, dz)
        for x in x_positions:
            for z in z_positions:
                self.set_recording_point(x, y, z, method=method)

    def set_recording_volume(
        self,
        x_min,
        x_max,
        y_min,
        y_max,
        z_min,
        z_max,
        dx=10,
        dy=10,
        dz=10,
        method="PSA",
    ):
        """
        Generate equaly spaced recording points in the y plane

        Parameters
        ----------
        x_min   : float
            minimal x postion for recording points, in um
        x_max   : float
            maximal x postion for recording points, in um
        y_min   : float
            minimal y postion for recording points, in um
        y_max   : float
            maximal y postion for recording points, in um
        z_min   : float
            minimal z postion for recording points, in um
        z_max   : float
            maximal z postion for recording points, in um
        dx      : float
            distance between recording points on the x coordinate, in um
        dy      : float
            distance between recording points on the y coordinate, in um
        dz      : float
            distance between recording points on the z coordinate, in um
        method  : string
            electrical potential approximation method, can be "PSA" (Point Source Approximation)
            or "LSA" (Line Source Approximation).
            set to "PSA" by default. Note that if LSA is requested with an anisotropic material, computation
            will automatically be performed using "PSA"
        """
        if np.abs(x_max - x_min) < dx:
            dx = np.abs(x_max - x_min) / 10
            rise_warning("dx too large, changed automatically to " + str(dx) + " um")
        if np.abs(y_max - y_min) < dy:
            dy = np.abs(y_max - y_min) / 10
            rise_warning("dy too large, changed automatically to " + str(dy) + " um")
        if np.abs(z_max - z_min) < dz:
            dz = np.abs(z_max - z_min) / 10
            rise_warning("dz too large, changed automatically to " + str(dz) + " um")
        x_positions = np.arange(x_min, x_max, dx)
        y_positions = np.arange(y_min, y_max, dy)
        z_positions = np.arange(z_min, z_max, dz)
        for x in x_positions:
            for y in y_positions:
                for z in z_positions:
                    self.set_recording_point(x, y, z, method=method)

    def compute_footprints(self, x_axon, y_axon, z_axon, d, ID, myelinated=False):
        """
        compute all footprints for a given axon on all recording points of the recorder

        Parameters
        ----------
        x_axon      : array
            array of the positions of the axon along the x-axis at which there is a current source, in [um]
        y_axon      : float
            y-axis position of the axon, in [um]
        z-axon      : float
            z-axis position of the axon, in [um]
        d           : float
            diameter of the axon, in [um]
        ID          : int
            axon ID number
        """
        if not self.is_empty():
            for point in self.recording_points:
                if myelinated:
                    if point.method == "PSA":
                        if self.isotropic:
                            point.compute_PSA_isotropic_footprint(
                                x_axon,
                                y_axon,
                                z_axon,
                                d,
                                ID,
                                self.sigma,
                                True,
                            )
                        else:
                            point.compute_PSA_anisotropic_footprint(
                                x_axon,
                                y_axon,
                                z_axon,
                                d,
                                ID,
                                self.sigma_xx,
                                self.sigma_yy,
                                self.sigma_zz,
                            )
                    if point.method == "LSA":
                        if self.isotropic:
                            point.compute_LSA_isotropic_footprint(
                                x_axon,
                                y_axon,
                                z_axon,
                                d,
                                ID,
                                self.sigma,
                                True,
                            )
                        else:
                            rise_warning(
                                "LSA not implemented for anisotropic material return isotropic by default"
                            )
                            point.compute_LSA_isotropic_footprint(
                                x_axon, y_axon, z_axon, d, ID, self.sigma
                            )
                else:
                    surface = np.pi * (d / cm) * (np.gradient(x_axon) / cm)
                    if point.method == "PSA":
                        if self.isotropic:
                            point.compute_PSA_isotropic_footprint(
                                x_axon,
                                y_axon,
                                z_axon,
                                d,
                                ID,
                                self.sigma,
                            )
                        else:
                            point.compute_PSA_anisotropic_footprint(
                                x_axon,
                                y_axon,
                                z_axon,
                                d,
                                ID,
                                self.sigma_xx,
                                self.sigma_yy,
                                self.sigma_zz,
                            )
                    if point.method == "LSA":
                        if self.isotropic:
                            point.compute_LSA_isotropic_footprint(
                                x_axon, y_axon, z_axon, d, ID, self.sigma
                            )
                        else:
                            rise_warning(
                                "LSA not implemented for anisotropic material return isotropic by default"
                            )
                            point.compute_LSA_isotropic_footprint(
                                x_axon, y_axon, z_axon, d, ID, self.sigma
                            )

    def init_recordings(self, N_points):
        """
        Initializes the recorded extracellular potential for all recodgin points.
        If a potential already exists,nothing will be performed

        Parameters
        ----------
        N_points : int
            length of the extracellular potential vector along temporal dimension
        """
        if not self.is_empty():
            for point in self.recording_points:
                point.init_recording(N_points)

    def reset_recordings(self, N_points):
        """
        Sets the recorded extracellular potential to zero whatever the conditions for all
        recording points

        Parameters
        ----------
        N_points : int
            length of the extracellular potential vector along temporal dimension
        """
        if not self.is_empty():
            for point in self.recording_points:
                point.reset_recording(N_points)

    def add_axon_contribution(self, I_membrane, ID):
        """
        Add one axon contribution to the extracellular potential of all recording points.

        Parameters
        ----------
        I_membrane  : array
            membrane current as a matrix over times (columns) and position (lines)
            in mA/cm2
        ID          : int
            axon ID as footprint are stored in a dictionnary with their ID as key
        """
        if not self.is_empty():
            for point in self.recording_points:
                point.add_axon_contribution(I_membrane, ID)

    def __update_master_t(self):
        """
        if the recorder has not been used in the master process:
        recover the time verctor in the master from an other process (rank 1)
        """
        synchronize_processes()
        if MCH.do_master_only_work():
            self.t = MCH.recieve_data_from_slave()["t"]
        elif MCH.rank == 1:
            MCH.send_data_to_master({"t": self.t})
        else:
            pass

    def gather_all_recordings(self, master_computed=True):
        """
        Gather all recordings computed by each cores in case of parallel simulation (fascicle
        level), sum de result and propagate final extracellular potential to each core.
        """
        if not self.is_empty():
            if not master_computed:
                self.__update_master_t()

            for point in self.recording_points:
                if point.recording is None:
                    point.recording = 0.0
                point.recording = MCH.sum_jobs(point.recording)



"""
NRV-I/O File Handler
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import json
import os

import ezdxf
import numpy as np

from .log_interface import rise_error, rise_warning


#################
# Miscalleneous #
#################
def is_iterable(some_stuff):
    """
    this function chels wether or not a variable contains an iterrable

    Parameters
    ----------
    some_stuff  :
        variable to check

    Returns
    -------
    False if a string or a number, True if iterrable (table, dict, tupple, numpy array...)
    """
    try:
        _ = (a for a in some_stuff)
        if isinstance(some_stuff, str):
            flag = False
        else:
            flag = True
    except TypeError:
        flag = False
    return flag


def rmv_ext(fname):
    """
    return filename without extension

    Parameters
    ----------
    fname   : str
        file name with or without extention

    Returns
    -------
    fname   : str
        file name without extention
    """
    if isinstance(fname, str):
        i = fname.rfind(".")
        if i > 0:
            fname = fname[:i]
    return fname


#####################################
## Folder and archive related code ##
#####################################
def create_folder(foldername, access_rights=0o755):
    """
    create a folder with controled access rights.

    Parameters
    ----------
    foldername : str
        name of the folder to create
    access_rights : int
        unix like rights
    """
    try:
        os.mkdir(foldername, access_rights)
    except OSError:
        rise_warning(
            "Creation of the directory %s failed, this folder may already exist"
            % foldername
        )


#######################
## JSON related code ##
#######################
def check_json_fname(fname):
    """
    Add ".json" extension is missing at the end of the file name and check if it exists.

    Parameters
    ----------
    fname    : str
        name of the file

    Retruns
    -------
    fname    : str
        name of the file with the ".json" extension added if required

    Errors
    ------
    NRV_Error
        rised if fname does not exist
    """
    if fname[-5:] != ".json":
        fname += ".json"
    if os.path.isfile(fname):
        return fname
    else:
        rise_error(fname + " not found cannot be load")


def json_dump(results, filename):
    """
    save stuff as a json file

    Parameters
    ----------
    results     :
        stuff to save
    filename    : str
        name of the file where results are saved
    """
    with open(filename, "w") as file_to_save:
        json.dump(results, file_to_save, cls=NRV_Encoder)


def json_load(filename):
    """
    Load stuff from a json file

    Parameters
    ----------
    filename    : str
        name of the file where results are stored

    Returns
    -------
    results : dictionary
        stuff from file
    """
    with open(check_json_fname(filename), "r") as file_to_read:
        results = json.load(file_to_read)
    return results


class NRV_Encoder(json.JSONEncoder):
    """
    Json encoding class, specific for NRV2 axon
    prevent from type error due to np.arrays
    solution taken as this from askpython.com
    """

    def default(self, obj):
        # If the object is a numpy array
        if isinstance(obj, np.integer):
            result = int(obj)
        elif isinstance(obj, np.floating):
            result = float(obj)
        elif isinstance(obj, np.ndarray):
            result = obj.tolist()
        else:
            # Let the base class Encoder handle the object
            result = json.JSONEncoder.default(self, obj)
        return result


######################
## DXF related code ##
######################
def load_dxf_file(filename):
    """
    UNDER DEV
    """
    doc = None
    try:
        doc = ezdxf.readfile(filename)
    except IOError:
        rise_error("Not a DXF file or a generic I/O error.")
    except ezdxf.DXFStructureError:
        rise_error("Invalid or corrupted DXF file.", out=2)
    return doc



"""
NRV-FEM
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from abc import abstractmethod

from ...backend.MCore import MCH
from ...backend.log_interface import pass_info, rise_warning
from ...backend.NRV_Class import NRV_class

###############
## Constants ##
###############
fem_verbose = True


###################
## Model classes ##
###################
class FEM_model(NRV_class):
    """
    A generic class for Finite Element models
    """

    @abstractmethod
    def __init__(self, Ncore=None):
        """
        Creates a instance of FEM_model

        Parameters
        ----------
        Ncore   : int
            number of cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        """
        super().__init__()
        self.Ncore = Ncore
        self.is_multi_proc = False
        self.type = "FEM"

        # Timmers
        self.meshing_timer = 0
        self.preparing_timer = 0
        self.solving_timer = 0
        self.access_res_timer = 0

        # Flags
        self.is_meshed = False
        self.is_computed = False

    def get_timers(self, verbose=False):
        """
        Returns the timers of each step of the FEM computation

        Parameters
        ----------
        verbose     : bool
            if true, timer are also passed to the log and the terminal (if verbosity level > 2)

        Returns
        -------
        meshing_timer   : float
            time spent to mesh the FEM problem, in s
        preparing_timer   : float
            time spent to mesh the prepare (i.e. generating variab connecting to the server), in s
        solving_timer   : float
            time spent to mesh the FEM problem, in s
        total_timer   : float
            time spent to mesh the FEM problem, in s
        """
        pass_info("mesh done in " + str(self.meshing_timer) + " s")
        pass_info("simulation prepared in " + str(self.preparing_timer) + " s")
        pass_info("simulation solved in " + str(self.solving_timer) + " s")
        pass_info("Time spent to access results " + str(self.access_res_timer) + " s")
        total_timer = (
            self.meshing_timer
            + self.preparing_timer
            + self.solving_timer
            + self.access_res_timer
        )
        pass_info("total duration " + str(total_timer) + " s")
        return self.meshing_timer, self.preparing_timer, self.solving_timer, total_timer

    def set_Ncore(self, N):
        """
        Set the number of cores to use for the FEM

        Parameters
        ----------
        N       : int
            Number of cores to set
        """
        self.Ncore = N
        if self.Ncore > MCH.size:
            rise_warning(
                "MCH size is " + str(MCH.size) + " cannot set FEM on more processes"
            )
            self.Ncore = MCH.size
        self.is_multi_proc = self.Ncore > 1



"""
NRV-myelinated
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import math

import numpy as np

from ..backend.log_interface import rise_warning
from .axons import *

MRG_fiberD = np.asarray([1, 2, 5.7, 7.3, 8.7, 10.0, 11.5, 12.8, 14.0, 15.0, 16.0])
MRG_g = np.asarray(
    [0.565, 0.585, 0.605, 0.630, 0.661, 0.690, 0.700, 0.719, 0.739, 0.767, 0.791]
)
MRG_axonD = np.asarray([0.8, 1.6, 3.4, 4.6, 5.8, 6.9, 8.1, 9.2, 10.4, 11.5, 12.7])
MRG_nodeD = np.asarray([0.7, 1.4, 1.9, 2.4, 2.8, 3.3, 3.7, 4.2, 4.7, 5.0, 5.5])
MRG_paraD1 = np.asarray([0.7, 1.4, 1.9, 2.4, 2.8, 3.3, 3.7, 4.2, 4.7, 5.0, 5.5])
MRG_paraD2 = np.asarray([0.8, 1.6, 3.4, 4.6, 5.8, 6.9, 8.1, 9.2, 10.4, 11.5, 12.7])
MRG_deltax = np.asarray([100, 200, 500, 750, 1000, 1150, 1250, 1350, 1400, 1450, 1500])
MRG_paralength2 = np.asarray([5, 10, 35, 38, 40, 46, 50, 54, 56, 58, 60])
MRG_nl = np.asarray([15, 20, 80, 100, 110, 120, 130, 135, 140, 145, 150])


def get_MRG_parameters(diameter):
    """
    Compute the MRG parameters

    Original parameters are listed and used if the input diameter is in the list,
    parameters are interpolated if the diameter is not in the original list.

    Attributes
    ----------
    diameter    : float
        diameter of the unmylinated axon to implement, in um

    Returns
    -------
    g           : float
    axonD       : float
    nodeD       : float
    paraD1      : float
    paraD2      : float
    deltax      : float
    paralength2 : float
    nl          : float
    """
    if diameter in MRG_fiberD:
        index = np.where(MRG_fiberD == diameter)[0]
        g = MRG_g[index]
        axonD = MRG_axonD[index]
        nodeD = MRG_nodeD[index]
        paraD1 = MRG_paraD1[index]
        paraD2 = MRG_paraD2[index]
        deltax = MRG_deltax[index]
        paralength2 = MRG_paralength2[index]
        nl = MRG_nl[index]
    else:
        # create fiting polynomyals
        g_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_g, 3))
        axonD_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_axonD, 3))
        nodeD_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_nodeD, 3))
        paraD1_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paraD1, 3))
        paraD2_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paraD2, 3))
        deltax_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_deltax, 4))
        if diameter < 1.0 or diameter > 16.0:
            # outside of the MRG originla limit, take 1st order approx,
            paralength2_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paralength2, 1))
        else:
            # try to fit a bit better
            paralength2_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paralength2, 5))
        nl_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_nl, 3))
        # evaluate for the requested diameter
        g = g_poly(diameter)
        axonD = axonD_poly(diameter)
        nodeD = nodeD_poly(diameter)
        paraD1 = paraD1_poly(diameter)
        paraD2 = paraD2_poly(diameter)
        deltax = deltax_poly(diameter)
        paralength2 = paralength2_poly(diameter)
        nl = nl_poly(diameter)
    return (
        float(g),
        float(axonD),
        float(nodeD),
        float(paraD1),
        float(paraD2),
        float(deltax),
        float(paralength2),
        float(nl),
    )


def get_length_from_nodes(diameter, nodes):
    """
    Function to compute the length of a myelinated axon to get the correct number of Nodes of Ranvier
    For Myelinated models only (not compatible with A delta thin myelinated models)

    Attributes
    ----------
    diameter    : float
        diameter of the axon in um
    nodes       : int
        number of nodes in the axon

    Returns
    -------
    length      : float
        lenth of the axon with the correct number of nodes in um
    """
    if diameter in MRG_fiberD:
        index = np.where(MRG_fiberD == diameter)[0]
        deltax = MRG_deltax[index]
    else:
        deltax_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_deltax, 4))
        deltax = deltax_poly(diameter)
    return float(math.ceil(deltax * (nodes - 1)))


class myelinated(axon):
    """
    Myelineated axon class. Automatic refinition of all neuron sections and properties. User-friendly object including model definition
    Inherit from axon class. see axon for further detail.
    """

    def __init__(
        self,
        y=0,
        z=0,
        d=10,
        L=10000,
        model="MRG",
        dt=0.001,
        node_shift=0,
        Nseg_per_sec=0,
        freq=100,
        freq_min=0,
        mesh_shape="plateau_sigmoid",
        alpha_max=0.3,
        d_lambda=0.1,
        rec="nodes",
        v_init=None,
        T=None,
        ID=0,
        threshold=-40,
        **kwargs,
    ):
        """
        initialisation of a myelinted axon

        Parameters
        ----------
        y               : float
            y coordinate for the axon, in um
        z               : float
            z coordinate for the axon, in um
        d               : float
            axon diameter, in um
        L               : float
            axon length along the x axins, in um
        model           : str
            choice of conductance double-cable based model, possibly:
                "MRG"           : see [1] for details
                "Gaines_motor"  : Gaines motor model, see [2]
                "Gaines_sensory": Gaines sensory model, see [2]
        dt              : float
            computation step for simulations, in ms. By default equal to 1 us
        node_shift      : float
            shift of the first node of Ranvier to zeros, as a fraction of internode length (0<= node_shift < 1)
        Nseg_per_sec    : int
            Number of segment per section in the axon. If set to 0, the number of segment is automatically computed using d-lambda rule and following paramters. If set by user, please use odd numbers
        freq            : float
            Frequency used for the d-lmbda rule, corresponding to the maximum membrane current frequency, by default set to 100 Hz
        freq_min        : float
            Minimal frequency fot the d-lambda rule when using an irregular number of segment along the axon, if set to 0, all sections have the same frequency determined by the previous parameter
        mesh_shape      : str
            Shape of the frequencial distribution for the dlmabda rule along the axon, pick between:
                "pyramidal"         -> min frequencies on both sides and linear increase up to the middle at the maximum frequency
                "sigmoid"           -> same a befor with sigmoid increase instead of linear
                "plateau"           -> sale as pyramidal except the max frequency is holded on a central plateau
                "plateau_sigmoid"   -> same as previous with sigmoid increase
        alpha_max       : float
            Proportion of the axon set to the maximum frequency for plateau shapes, by default set to 0.3
        d_lambda        : float
            value of d-lambda for the dlambda rule
        rec             : str
            recording zones for the membrane potential, eiter:
                "nodes" -> record only at the nodes of Ranvier
            or
                "all" -> all computation points in nodes of Ranvier and over myelin
        v_init          : float
            Initial value of the membrane voltage in mV, set None to get an automatically model attributed value
        T               : gloat
            temperature in C, set None to get an automatically model attributed value
        ID              : int
            axon ID, by default set to 0,
        threshold       : float
            voltage threshold in mV for further spike detection in post-processing, by defautl set to -40mV, see post-processing library for further help

        Note
        ----
        scientific sources for models:
        [1] McIntyre CC, Richardson AG, and Grill WM. Modeling the excitability of mammalian nerve fibers: influence of afterpotentials on the recovery cycle. Journal of Neurophysiology 87:995-1006, 2002.
        [2] Gaines, J. L., Finn, K. E., Slopsema, J. P., Heyboer, L. A.,  Polasek, K. H. (2018). A model of motor and sensory axon activation in the median nerve using surface electrical stimulation. Journal of computational neuroscience, 45(1), 29-43.
        """
        super().__init__(
            y,
            z,
            d,
            L,
            dt=dt,
            Nseg_per_sec=Nseg_per_sec,
            freq=freq,
            freq_min=freq_min,
            mesh_shape=mesh_shape,
            alpha_max=alpha_max,
            d_lambda=d_lambda,
            v_init=v_init,
            T=T,
            ID=ID,
            threshold=threshold,
            **kwargs,
        )
        self.myelinated = True
        self.thin = False
        self.rec = rec
        self.node_shift = node_shift

        if model in myelinated_models:
            self.model = model
        else:
            self.model = "MRG"
        self.__compute_axon_parameters()

    def __compute_axon_parameters(self):
        """
        generate axon from parameters set by user
        """
        if self.v_init is None:
            # model driven
            if self.model == "Gaines_sensory":
                self.v_init = -79.3565
            elif self.model == "Gaines_motor":
                self.v_init = -85.9411
            else:
                self.v_init = -80
        else:
            # user driven
            self.v_init = self.v_init
        ## Handling temperature
        if self.T is None:
            # model driven
            self.T = 37

        ############################################
        ## PARMETERS FOR THE COMPARTIMENTAL MODEL ##
        ############################################
        # compute variable MRG parameters, (usefull also if non MRG models)
        (
            self.g,
            self.axonD,
            self.nodeD,
            self.paraD1,
            self.paraD2,
            self.deltax,
            self.paralength2,
            self.nl,
        ) = get_MRG_parameters(self.d)
        # Morphological parameters
        self.paralength1 = 3
        self.nodelength = 1.0
        self.space_p1 = 0.002
        self.space_p2 = 0.004
        self.space_i = 0.004
        self.interlength = (
            self.deltax
            - self.nodelength
            - (2 * self.paralength1)
            - (2 * self.paralength2)
        ) / 6
        # electrical parameters
        self.rhoa = 0.7e6  # Ohm-um
        self.mycm = 0.1  # uF/cm2/lamella membrane
        self.mygm = 0.001  # S/cm2/lamella membrane
        self.Rpn0 = (self.rhoa * 0.01) / (
            math.pi
            * ((((self.nodeD / 2) + self.space_p1) ** 2) - ((self.nodeD / 2) ** 2))
        )
        self.Rpn1 = (self.rhoa * 0.01) / (
            math.pi
            * ((((self.paraD1 / 2) + self.space_p1) ** 2) - ((self.paraD1 / 2) ** 2))
        )
        self.Rpn2 = (self.rhoa * 0.01) / (
            math.pi
            * ((((self.paraD2 / 2) + self.space_p2) ** 2) - ((self.paraD2 / 2) ** 2))
        )
        self.Rpx = (self.rhoa * 0.01) / (
            math.pi
            * ((((self.axonD / 2) + self.space_i) ** 2) - ((self.axonD / 2) ** 2))
        )
        #########################
        ## morphology planning ##
        #########################
        # basic MRG sequence
        self.MRG_Sequence = [
            "node",
            "MYSA",
            "FLUT",
            "STIN",
            "STIN",
            "STIN",
            "STIN",
            "STIN",
            "STIN",
            "FLUT",
            "MYSA",
        ]
        # basic MRG starts with the node, if needed, adapt the sequence
        if self.node_shift == 0 or self.node_shift == 1:
            # no Rotation
            self.this_ax_sequence = self.MRG_Sequence
            self.first_section_size = self.nodelength

        elif self.node_shift < (self.paralength1) / self.deltax:
            # rotation of less than 1 MYSA
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 1)
            self.first_section_size = self.node_shift * self.deltax

        elif self.node_shift < (self.paralength1 + self.paralength2) / self.deltax:
            # rotation of 1 MYSA and less than one FLUT
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 2)
            self.first_section_size = self.node_shift * self.deltax - self.paralength1

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 3)
            self.first_section_size = (
                self.node_shift * self.deltax - self.paralength1 - self.paralength2
            )

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + 2 * self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 1 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 4)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - self.interlength
            )

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + 3 * self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 2 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 5)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - 2 * self.interlength
            )

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + 4 * self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 3 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 6)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - 3 * self.interlength
            )

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + 5 * self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 4 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 7)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - 4 * self.interlength
            )

        elif (
            self.node_shift
            < (self.paralength1 + self.paralength2 + 6 * self.interlength) / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 5 STIN and less than a STIN
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 8)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - 5 * self.interlength
            )

        elif (
            self.node_shift
            < (self.paralength1 + 2 * self.paralength2 + 6 * self.interlength)
            / self.deltax
        ):
            # rotation of 1 MYSA, 1 FLUT, 6 STIN and less than a FLUT
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 9)
            self.first_section_size = (
                self.node_shift * self.deltax
                - self.paralength1
                - self.paralength2
                - 6 * self.interlength
            )

        else:
            # rotation of 1 MYSA, 2 FLUT, 6 STIN and less than a MYSA
            # WARNING FOR DEV : the unprobable case of a node cut in two halfs is not considered...
            self.this_ax_sequence = rotate_list(self.MRG_Sequence, 10)
            self.first_section_size = self.deltax * (1 - self.node_shift)
        ################
        ## morphology ##
        ################
        self.axonnodes = 0  # number of nodes in the axon
        self.node = []  # list of nodes in the axon
        self.paranodes1 = 0  # number of MYSA in the axon
        self.MYSA = []  # list of MYSA in the axon
        self.paranodes2 = 0  # number of FLUT in the axon
        self.FLUT = []  # list of FLUT in the axon
        self.axoninter = 0  # number of STIN in the axon
        self.STIN = []  # list of STIN in the axon

        prov_length = 0
        self.Nsec = 0
        self.last_section_size = 0
        self.last_section_kind = None
        self.axon_path_type = []
        self.axon_path_index = []
        while prov_length < self.L:
            # axon is too short, create a section
            pos_in_sequence = (self.Nsec) % len(
                self.this_ax_sequence
            )  # find the current position in the node sequence
            if self.this_ax_sequence[pos_in_sequence] == "node":
                # you need to create a node
                self.node.append(neuron.h.Section(name="node[%d]" % self.axonnodes))
                self.axon_path_type.append("node")
                self.axon_path_index.append(self.axonnodes)
                self.axonnodes += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.nodelength >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = "node"
                    self.last_section_size = self.L - prov_length
                prov_length += self.nodelength
                # CONNECT THE NODE TO ITS PARENT: a node can only have a MYSA for parent, check it"s not the start...
                if self.paranodes1 != 0:
                    self.node[-1].connect(self.MYSA[-1], 1, 0)

            elif self.this_ax_sequence[pos_in_sequence] == "MYSA":
                # you need to create a MYSA
                self.MYSA.append(neuron.h.Section(name="MYSA[%d]" % self.paranodes1))
                self.axon_path_type.append("MYSA")
                self.axon_path_index.append(self.paranodes1)
                self.paranodes1 += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.paralength1 >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = "MYSA"
                    self.last_section_size = self.L - prov_length
                prov_length += self.paralength1
                # CONNECT THE MYSA TO ITS PARENT: a MYSA can have for parent a node or a FLUT, check it"s not the start...
                connect_to_type = self.this_ax_sequence[
                    pos_in_sequence - 1
                ]  # get parent type
                if connect_to_type == "node" and self.axonnodes != 0:
                    self.MYSA[-1].connect(self.node[-1], 1, 0)
                if connect_to_type == "FLUT" and self.paranodes2 != 0:
                    self.MYSA[-1].connect(self.FLUT[-1], 1, 0)

            elif self.this_ax_sequence[pos_in_sequence] == "FLUT":
                # you need to create a FLUT
                self.FLUT.append(neuron.h.Section(name="FLUT[%d]" % self.paranodes2))
                self.axon_path_type.append("FLUT")
                self.axon_path_index.append(self.paranodes2)
                self.paranodes2 += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.paralength2 >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = "FLUT"
                    self.last_section_size = self.L - prov_length
                prov_length += self.paralength2
                # CONNECT THE FLUT TO ITS PARENT: a FLUT can have for parent a MYSA of a STIN, check it"s not the start...
                connect_to_type = self.this_ax_sequence[
                    pos_in_sequence - 1
                ]  # get parent type
                if connect_to_type == "MYSA" and self.paranodes1 != 0:
                    self.FLUT[-1].connect(self.MYSA[-1], 1, 0)
                if connect_to_type == "STIN" and self.axoninter != 0:
                    self.FLUT[-1].connect(self.STIN[-1], 1, 0)

            else:
                # you need to create a STIN
                self.STIN.append(neuron.h.Section(name="STIN[%d]" % self.axoninter))
                self.axon_path_type.append("STIN")
                self.axon_path_index.append(self.axoninter)
                self.axoninter += 1
                self.Nsec += 1
                # increment the prov length and check it is not to much
                if prov_length + self.interlength >= self.L:
                    # it will be the last section to add to the axon
                    self.last_section_kind = "STIN"
                    self.last_section_size = self.L - prov_length
                prov_length += self.interlength
                # CONNECT THE STIN TO ITS PARENT: a STIN can have for parent a STIN od a FLUT, check it"s not the start...
                connect_to_type = self.this_ax_sequence[pos_in_sequence - 1]
                if connect_to_type == "STIN" and self.axoninter > 1:
                    self.STIN[-1].connect(self.STIN[-2], 1, 0)
                if connect_to_type == "FLUT" and self.paranodes2 != 0:
                    self.STIN[-1].connect(self.FLUT[-1], 1, 0)

        if self.node == []:
            rise_warning(
                "Warning, myelinated axon without node... this can cause latter errors and is maybe unwanted ?\n"
            )
            # logging.warning("Warning, myelinated axon without node... this can cause latter errors and is maybe unwanted ?\n")

        ####################
        ## programm model ##
        ####################
        self.__set_model(self.model)
        # adjust the length of the first section
        if self.this_ax_sequence[0] == "node":
            self.node[0].L = self.first_section_size
        elif self.this_ax_sequence[0] == "MYSA":
            self.MYSA[0].L = self.first_section_size
        elif self.this_ax_sequence[0] == "FLUT":
            self.FLUT[0].L = self.first_section_size
        else:  # should be a STIN
            self.STIN[0].L = self.first_section_size
        # adjust the length if the last section
        if self.last_section_kind == "node":
            self.node[-1].L = self.last_section_size
        elif self.last_section_kind == "MYSA":
            self.MYSA[-1].L = self.last_section_size
        elif self.last_section_kind == "FLUT":
            self.FLUT[-1].L = self.last_section_size
        else:  # should be a STIN
            self.STIN[-1].L = self.last_section_size
        # define the geometry of the axon
        self._axon__define_shape()
        # define the number of segments
        self.__set_Nseg()
        # get nodes positions
        self.__get_seg_positions()
        self.__get_rec_positions()

    def save(
        self,
        save=False,
        fname="axon.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Return axon as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "axon.json"

        Returns
        -------
        ax_dic : dict
            dictionary containing all information
        """
        blacklist += ["node", "MYSA", "FLUT", "STIN"]
        return super().save(
            save=save,
            fname=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
            blacklist=blacklist,
        )

    def __set_model(self, model):
        """
        Set the double cable model. For internal use only.

        Parameters
        ----------
        model           : str
            choice of conductance double-cable based model, possibly:
                "MRG"           : see [1] for details
                "Gaines_motor"  : Gaines motor model, see [2]
                "Gaines_sensory": Gaines sensory model, see [2]

        Note
        ----
        scientific sources for models:
        [1] McIntyre CC, Richardson AG, and Grill WM. Modeling the excitability of mammalian nerve fibers: influence of afterpotentials on the recovery cycle. Journal of Neurophysiology 87:995-1006, 2002.
        [2] Gaines, J. L., Finn, K. E., Slopsema, J. P., Heyboer, L. A.,  Polasek, K. H. (2018). A model of motor and sensory axon activation in the median nerve using surface electrical stimulation. Journal of computational neuroscience, 45(1), 29-43.
        """
        for n in self.node:
            n.nseg = 1
            n.diam = self.nodeD
            n.L = self.nodelength
            n.Ra = self.rhoa / 10000
            n.cm = 2
            if model == "Gaines_sensory":
                n.insert("node_sensory")
            elif model == "Gaines_motor":
                n.insert("node_motor")
            else:
                n.insert("axnode")
            n.insert("extracellular")
            n.xraxial[0] = self.Rpn0
            n.xg[0] = 1e10
            n.xc[0] = 0
        for m in self.MYSA:
            m.nseg = 1
            m.diam = self.d
            m.L = self.paralength1
            m.Ra = self.rhoa * (1 / (self.paraD1 / self.d) ** 2) / 10000
            m.cm = 2 * self.paraD1 / self.d
            if model == "Gaines_sensory":
                m.insert("mysa_sensory")
            elif model == "Gaines_motor":
                m.insert("mysa_motor")
            else:
                m.insert("pas")
                m.g_pas = 0.001 * self.paraD1 / self.d
                m.e_pas = -80
            m.insert("extracellular")
            m.xraxial[0] = self.Rpn1
            m.xg[0] = self.mygm / (self.nl * 2)
            m.xc[0] = self.mycm / (self.nl * 2)
        for f in self.FLUT:
            f.nseg = 1
            f.diam = self.d
            f.L = self.paralength2
            f.Ra = self.rhoa * (1 / (self.paraD2 / self.d) ** 2) / 10000
            f.cm = 2 * self.paraD2 / self.d
            if model == "Gaines_sensory":
                f.insert("flut_sensory")
            elif model == "Gaines_motor":
                f.insert("flut_motor")
            else:
                f.insert("pas")
                f.g_pas = 0.0001 * self.paraD2 / self.d
                f.e_pas = -80
            f.insert("extracellular")
            f.xraxial[0] = self.Rpn2
            f.xg[0] = self.mygm / (self.nl * 2)
            f.xc[0] = self.mycm / (self.nl * 2)
        for s in self.STIN:
            s.nseg = 1
            s.diam = self.d
            s.L = self.interlength
            s.Ra = self.rhoa * (1 / (self.axonD / self.d) ** 2) / 10000
            s.cm = 2 * self.axonD / self.d
            if model == "Gaines_sensory":
                s.insert("stin_sensory")
            elif model == "Gaines_motor":
                s.insert("stin_motor")
            else:
                s.insert("pas")
                s.g_pas = 0.0001 * self.axonD / self.d
                s.e_pas = -80
            s.insert("extracellular")
            s.xraxial[0] = self.Rpx
            s.xg[0] = self.mygm / (self.nl * 2)
            s.xc[0] = self.mycm / (self.nl * 2)

    def set_Markov_Nav(self, list_of_nodes=[]):
        """
        Change typical particle-Na sodium in Hodgking-Huxley formalism to Markov-channel population model.

        Parameters
        ----------
        list_of_nodes   : list, array, np.array
            list of Nodes of Ranier to modify, if empty, all nodes sodium channels are changed

        Note
        ----
        based on:
        Yi, G., and Grill, W. M. (2020). Kilohertz waveforms optimized to produce closed-state Na+ channel inactivation eliminate onset response in nerve conduction block. PLoS computational biology, 16(6), e1007766.

        !!! TO USE WITH CAUTION !!!
        """
        if list_of_nodes == []:
            list_of_nodes = np.arange(self.axonnodes)
        if self.model == "MRG":
            for NoR in list_of_nodes:
                self.node[NoR].gnapbar_axnode = 0
                self.node[NoR].gnabar_axnode = 0
        elif self.model == "Gaines_motor":
            for NoR in list_of_nodes:
                self.node[NoR].gnapbar_node_motor = 0
                self.node[NoR].gnabar_node_motor = 0
        elif self.model == "Gaines_sensory":
            for NoR in list_of_nodes:
                self.node[NoR].gnapbar_node_sensory = 0
                self.node[NoR].gnabar_node_sensory = 0
        # insert channel specific mechanisms
        # Nav1.1 models fast sodium
        # Nav1.6 models persistant sodium, to be confirmed, Warning gnabar is different than MRG and Gaines
        for NoR in list_of_nodes:
            # insert Nav1.1 model
            self.node[NoR].insert("na11a")
            self.node[NoR].gbar_na11a = 11.9
            # insert Nav1.6 model
            self.node[NoR].insert("na16a")
            self.node[NoR].gbar_na16a = 0.01
            ## WARINING -  Sodium Nernst potential is defined for axnode as ena_axnode, and is common
            ## for fast and persistant sodium channels (see lines 98 and 99 in AXNOE.mod)
            ## This is not the case for Nav1.x, which uses the NEURON instruction USEION na READ ena
            ## meaning the following line is tuning directly the na11a and na16a mechanisms
            self.node[NoR].ena = 50
        self.Markov_Nav_modeled_NoR = list_of_nodes

    def __set_Nseg(self):
        """
        Set the number of segments automatically acording initialization. For internal use only.
        """

        if type(self.Nseg_per_sec) is dict:
            for n in self.node:
                n.nseg = self.Nseg_per_sec["node"]
            for m in self.MYSA:
                m.nseg = self.Nseg_per_sec["MYSA"]
            for f in self.FLUT:
                f.nseg = self.Nseg_per_sec["FLUT"]
            for s in self.STIN:
                s.nseg = self.Nseg_per_sec["STIN"]
            self.node_Nseg = self.axonnodes * self.Nseg_per_sec["node"]
            self.MYSA_Nseg = self.paranodes1 * self.Nseg_per_sec["MYSA"]
            self.FLUT_Nseg = self.paranodes2 * self.Nseg_per_sec["FLUT"]
            self.STIN_Nseg = self.axoninter * self.Nseg_per_sec["STIN"]

        elif self.Nseg_per_sec != 0:
            # all sections will have a fixed number of segment chosen by user
            for n in self.node:
                n.nseg = self.Nseg_per_sec
            for m in self.MYSA:
                m.nseg = self.Nseg_per_sec
            for f in self.FLUT:
                f.nseg = self.Nseg_per_sec
            for s in self.STIN:
                s.nseg = self.Nseg_per_sec
            self.node_Nseg = self.axonnodes * self.Nseg_per_sec
            self.MYSA_Nseg = self.paranodes1 * self.Nseg_per_sec
            self.FLUT_Nseg = self.paranodes2 * self.Nseg_per_sec
            self.STIN_Nseg = self.axoninter * self.Nseg_per_sec
        else:
            # here comes the dlambda rule
            self.node_Nseg = 0
            self.MYSA_Nseg = 0
            self.FLUT_Nseg = 0
            self.STIN_Nseg = 0
            if self.freq_min == 0:
                # uniform meshing
                for n in self.node:
                    Nseg = d_lambda_rule(n.L, self.d_lambda, self.freq, n)
                    n.nseg = Nseg
                    self.node_Nseg += Nseg
                for m in self.MYSA:
                    Nseg = Nseg = d_lambda_rule(m.L, self.d_lambda, self.freq, m)
                    m.nseg = Nseg
                    self.MYSA_Nseg += Nseg
                for f in self.FLUT:
                    Nseg = Nseg = d_lambda_rule(f.L, self.d_lambda, self.freq, f)
                    f.nseg = Nseg
                    self.FLUT_Nseg += Nseg
                for s in self.STIN:
                    Nseg = Nseg = d_lambda_rule(s.L, self.d_lambda, self.freq, s)
                    s.nseg = Nseg
                    self.STIN_Nseg += Nseg
            else:
                # non-uniform meshing
                freqs = create_Nseg_freq_shape(
                    self.Nsec, self.mesh_shape, self.freq, self.freq_min, self.alpha_max
                )
                for k in range(len(self.axon_path_type)):
                    sec_type = self.axon_path_type[k]
                    sec_index = self.axon_path_index[k]
                    if sec_type == "node":
                        Nseg = d_lambda_rule(
                            self.node[sec_index].L,
                            self.d_lambda,
                            freqs[k],
                            self.node[sec_index],
                        )
                        self.node[sec_index].nseg = Nseg
                        self.node_Nseg += Nseg
                    elif sec_type == "MYSA":
                        Nseg = d_lambda_rule(
                            self.MYSA[sec_index].L,
                            self.d_lambda,
                            freqs[k],
                            self.MYSA[sec_index],
                        )
                        self.MYSA[sec_index].nseg = Nseg
                        self.MYSA_Nseg += Nseg
                    elif sec_type == "FLUT":
                        Nseg = d_lambda_rule(
                            self.FLUT[sec_index].L,
                            self.d_lambda,
                            freqs[k],
                            self.FLUT[sec_index],
                        )
                        self.FLUT[sec_index].nseg = Nseg
                        self.FLUT_Nseg += Nseg
                    else:  # should be STIN
                        Nseg = d_lambda_rule(
                            self.STIN[sec_index].L,
                            self.d_lambda,
                            freqs[k],
                            self.STIN[sec_index],
                        )
                        self.STIN[sec_index].nseg = Nseg
                        self.STIN_Nseg += Nseg
        self.Nseg = self.node_Nseg + self.MYSA_Nseg + self.FLUT_Nseg + self.STIN_Nseg

    def __get_seg_positions(self):
        """
        Get segment positions, for internal use only.
        """
        x_offset = 0
        x = []
        x_nodes = []
        nodes_index = []
        self.rec_position_list = []
        for k in range(len(self.axon_path_type)):
            sec_type = self.axon_path_type[k]
            sec_index = self.axon_path_index[k]
            self.rec_position_list.append([])
            if sec_type == "node":
                x_nodes.append(x_offset + self.node[sec_index].L / 2)
                for seg in self.node[sec_index].allseg():
                    if x == []:
                        x.append(seg.x * (self.node[sec_index].L) + x_offset)
                        nodes_index.append(0)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x * (self.node[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            nodes_index.append(len(x) - 1)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.node[sec_index].L
            elif sec_type == "MYSA":
                for seg in self.MYSA[sec_index].allseg():
                    if x == []:
                        x.append(seg.x * (self.MYSA[sec_index].L) + x_offset)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x * (self.MYSA[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.MYSA[sec_index].L
            elif sec_type == "FLUT":
                for seg in self.FLUT[sec_index].allseg():
                    if x == []:
                        x.append(seg.x * (self.FLUT[sec_index].L) + x_offset)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x * (self.FLUT[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.FLUT[sec_index].L
            else:  # should be STIN
                for seg in self.STIN[sec_index].allseg():
                    if x == []:
                        x.append(seg.x * (self.STIN[sec_index].L) + x_offset)
                        self.rec_position_list[-1].append(seg.x)
                    else:
                        x_seg = seg.x * (self.STIN[sec_index].L) + x_offset
                        if x_seg != x[-1]:
                            x.append(x_seg)
                            self.rec_position_list[-1].append(seg.x)
                x_offset += self.STIN[sec_index].L
        self.x = np.asarray(x)
        self.x_nodes = np.asarray(x_nodes)
        self.node_index = np.asarray(nodes_index)

    def __get_rec_positions(self):
        """
        Get the position of points with voltage recording. For internal use only.
        """
        if self.rec == "nodes":
            self.x_rec = self.x_nodes
        else:
            self.x_rec = self.x

    ###############################
    ## Intracellular stimulation ##
    ###############################
    def insert_I_Clamp_node(self, index, t_start, duration, amplitude):
        """
        Insert a IC clamp stimulation on a Ranvier node at its midd point position

        Parameters
        ----------
        index       : int
            node number of the node to stimulate
        t_start     : float
            starting time (ms)
        duration    : float
            duration of the pulse(ms)
        amplitude   : float
            amplitude of the pulse (nA)
        """
        # add the stimulation to the axon
        self.intra_current_stim.append(neuron.h.IClamp(0.5, sec=self.node[index]))
        # modify the stimulation parameters
        self.intra_current_stim[-1].delay = t_start
        self.intra_current_stim[-1].dur = duration
        self.intra_current_stim[-1].amp = amplitude
        # save the stimulation parameter for results
        self.intra_current_stim_positions.append(self.x_nodes[index])
        self.intra_current_stim_starts.append(t_start)
        self.intra_current_stim_durations.append(duration)
        self.intra_current_stim_amplitudes.append(amplitude)

    def insert_I_Clamp(self, position, t_start, duration, amplitude):
        """
        Insert a IC clamp stimulation at the midd point of the nearest node to the specified position

        Parameters
        ----------
        position    : float
            relative position over the axon
        t_start     : float
            starting time (ms)
        duration    :
            duration of the pulse(ms)
        amplitude   :
            amplitude of the pulse (nA)
        """
        # adapt position to the number of sections
        index = round((position * (self.axonnodes - 1) + 0.5))
        self.insert_I_Clamp_node(index, t_start, duration, amplitude)

    def insert_V_Clamp_node(self, index, stimulus):
        """
        Insert a V clamp stimulation

        Parameters
        ----------
        index       : int
            node number of the node to stimulate
        stimulus    : stimulus object
            stimulus for the clamp, see Stimulus.py for more information
        """
        # add the stimulation to the axon
        self.intra_voltage_stim = neuron.h.VClamp(0.5, sec=self.node[index])
        # save the stimulation parameter for results
        self.intra_voltage_stim_position.append(self.x_nodes[index])
        # save the stimulus for later use
        self.intra_voltage_stim_stimulus = stimulus
        # set fake duration
        self.intra_voltage_stim.dur[0] = 1e9

    def insert_V_Clamp(self, position, stimulus):
        """
        Insert a V clamp stimulation at the midd point of the nearest node to the specified position

        Parameters
        ----------
        position    : float
            relative position over the axon
        stimulus    : stimulus object
            stimulus for the clamp, see Stimulus.py for more information
        """
        # adapt position to the number of sections
        index = round((position * (self.axonnodes - 1) + 0.5))
        self.insert_V_Clamp_node(index, stimulus)

    ##############################
    ## Result recording methods ##
    ##############################
    def __set_recorders_with_key(
        self,
        reclist,
        key=None,
        single_mod=True,
        key_node=None,
        key_MYSA=None,
        key_FLUT=None,
        key_STIN=None,
    ):
        """
        To automate the methods set_recorder. For internal use only.
        Parameters
        ----------
        reclist     : neuron.h.List
            List in witch the reccorders should be saved
        key         : str
            Name of the key which should be recorded, if None result of the recording set to [0],
            by default None
        single_mod  : bool
            if True key will be used for all the section, else specific key will be used in each
            section, by default True
        key_node    : str
            Name of the key which should be recorded in node, if None result of the recording set
            to [0], by default None
        key_MYSA    : str
            Name of the key which should be recorded in MYSA, if None result of the recording set
            to [0], by default None
        key_FLUT    : str
            Name of the key which should be recorded in FLUT, if None result of the recording set
            to [0], by default None
        key_STIN    : str
            Name of the key which should be recorded in STIN, if None result of the recording set
            to [0], by default None
        """
        if single_mod:
            key_node, key_MYSA, key_FLUT, key_STIN = key, key, key, key

        if self.rec == "nodes":
            # recording only on middle of all nodes
            for n in self.node:
                if key_node is not None:
                    rec = neuron.h.Vector().record(getattr(n(0.5), key_node), sec=n)
                else:
                    rec = neuron.h.Vector([0])
                reclist.append(rec)
        else:
            # recording on all segments
            for k in range(len(self.axon_path_type)):
                sec_type = self.axon_path_type[k]
                sec_index = self.axon_path_index[k]
                for position in self.rec_position_list[k]:
                    if sec_type == "node":
                        if key_node is not None:
                            rec = neuron.h.Vector().record(
                                getattr(self.node[sec_index](position), key_node),
                                sec=self.node[sec_index],
                            )
                        else:
                            rec = neuron.h.Vector([0])
                        reclist.append(rec)
                    elif sec_type == "MYSA":
                        if key_MYSA is not None:
                            rec = neuron.h.Vector().record(
                                getattr(self.MYSA[sec_index](position), key_MYSA),
                                sec=self.MYSA[sec_index],
                            )
                        else:
                            rec = neuron.h.Vector([0])
                        reclist.append(rec)
                    elif sec_type == "FLUT":
                        if key_FLUT is not None:
                            rec = neuron.h.Vector().record(
                                getattr(self.FLUT[sec_index](position), key_FLUT),
                                sec=self.FLUT[sec_index],
                            )
                        else:
                            rec = neuron.h.Vector([0])
                        reclist.append(rec)
                    else:  # should be STIN
                        if key_STIN is not None:
                            rec = neuron.h.Vector().record(
                                getattr(self.STIN[sec_index](position), key_STIN),
                                sec=self.STIN[sec_index],
                            )
                        else:
                            rec = neuron.h.Vector([0])
                        reclist.append(rec)

    def __get_recorders_from_list(self, reclist):
        """
        Convert reclist in np.array To automate methods set_recorder. For internal use only.
        Parameters
        ----------
        reclist     : neuron.h.List
            List in witch the reccorders are saved

        Returns
        -------
        val         : np.array
            array of every recorded value for all rec point and time
        """
        if self.rec == "nodes":
            dim = (self.axonnodes, self.t_len)
        else:
            dim = (len(self.x_rec), self.t_len)

        val = np.zeros(dim)
        for k in range(dim[0]):
            val[k, :] = np.asarray(reclist[k])
        return val

    def __get_var_from_mod(
        self,
        key=None,
        single_mod=True,
        key_node=None,
        key_MYSA=None,
        key_FLUT=None,
        key_STIN=None,
    ):
        """
        return a column with value in every recording point of a constant from a mod. For internal use only.
        """
        if single_mod:
            key_node, key_MYSA, key_FLUT, key_STIN = key, key, key, key

        if self.rec == "nodes":
            val = np.zeros(len(self.node))
            if key_node is not None:
                val[:] = getattr(self.node[0](0.5), key_node)[0]
        else:
            val = np.zeros((len(self.x_rec)))
            i = -1
            for k in range(len(self.axon_path_type)):
                sec_type = self.axon_path_type[k]
                sec_index = self.axon_path_index[k]
                for position in self.rec_position_list[k]:
                    i += 1
                    if sec_type == "node":
                        if key_node is not None:
                            v = getattr(self.node[sec_index](position), key_node)[0]
                    elif sec_type == "MYSA":
                        if key_MYSA is not None:
                            v = getattr(self.MYSA[sec_index](position), key_MYSA)[0]
                    elif sec_type == "FLUT":
                        if key_FLUT is not None:
                            v = getattr(self.FLUT[sec_index](position), key_FLUT)[0]
                    else:  # should be STIN
                        if key_STIN is not None:
                            v = getattr(self.STIN[sec_index](position), key_STIN)[0]
                    if isinstance(v, float):
                        val[i] = v
                    else:   # needed for xc, xg, xraxial (see if it can be simplify)
                        val[i] = v[0]
        return val

    def set_membrane_voltage_recorders(self):
        """
        Prepare the membrane voltage recording. For internal use only.
        """
        self.v_reclist = neuron.h.List()
        key = "_ref_v"
        self.__set_recorders_with_key(self.v_reclist, key)

    def get_membrane_voltage(self):
        """
        get the membrane voltage at the end of simulation. For internal use only.
        """
        return self.__get_recorders_from_list(self.v_reclist)

    def set_membrane_current_recorders(self):
        """
        Prepare the membrane current recording. For internal use only.
        """
        self.i_reclist = neuron.h.List()
        key = "_ref_i_membrane"
        self.__set_recorders_with_key(self.i_reclist, key)

    def get_membrane_current(self):
        """
        get the membrane current at the end of simulation. For internal use only.
        """
        return self.__get_recorders_from_list(self.i_reclist)

    def set_ionic_current_recorders(self):
        """
        Prepare the ionic channels current recording. For internal use only.
        """
        if self.model == "MRG":
            self.ina_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.ina_reclist, single_mod=False, key_node="_ref_ina_axnode"
            )
            self.inap_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.inap_reclist, single_mod=False, key_node="_ref_inap_axnode"
            )
            self.ik_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.ik_reclist, single_mod=False, key_node="_ref_ik_axnode"
            )
            self.il_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.il_reclist, single_mod=False, key_node="_ref_il_axnode"
            )
        else:  # should be Gaines, motor or sensory
            if self.model == "Gaines_motor":
                key_mod = "_motor"
            else:
                key_mod = "_sensory"
            self.ina_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.ina_reclist, single_mod=False, key_node="_ref_ina_node" + key_mod
            )

            self.inap_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.inap_reclist, single_mod=False, key_node="_ref_inap_node" + key_mod
            )

            self.ik_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.ik_reclist,
                single_mod=False,
                key_node="_ref_ik_node" + key_mod,
                key_MYSA="_ref_ik_mysa" + key_mod,
                key_FLUT="_ref_ik_flut" + key_mod,
                key_STIN="_ref_ik_stin" + key_mod,
            )

            self.ikf_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.ikf_reclist,
                single_mod=False,
                key_node="_ref_ikf_node" + key_mod,
                key_MYSA="_ref_ikf_mysa" + key_mod,
                key_FLUT="_ref_ikf_flut" + key_mod,
                key_STIN="_ref_ikf_stin" + key_mod,
            )

            self.iq_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.iq_reclist,
                single_mod=False,
                key_MYSA="_ref_iq_mysa" + key_mod,
                key_FLUT="_ref_iq_flut" + key_mod,
                key_STIN="_ref_iq_stin" + key_mod,
            )

            self.il_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.il_reclist,
                single_mod=False,
                key_node="_ref_il_node" + key_mod,
                key_MYSA="_ref_il_mysa" + key_mod,
                key_FLUT="_ref_il_flut" + key_mod,
                key_STIN="_ref_il_stin" + key_mod,
            )

    def get_ionic_current(self):
        """
        get the ionic channels currents at the end of simulation. For internal use only.
        """
        results = []
        results += [self.__get_recorders_from_list(self.ina_reclist)]
        results += [self.__get_recorders_from_list(self.inap_reclist)]
        results += [self.__get_recorders_from_list(self.ik_reclist)]
        if self.model == "Gaines_motor" or self.model == "Gaines_sensory":
            results += [self.__get_recorders_from_list(self.ikf_reclist)]
            results += [self.__get_recorders_from_list(self.iq_reclist)]
        results += [self.__get_recorders_from_list(self.il_reclist)]
        return results

    def set_particules_values_recorders(self):
        """
        Prepare the particules current recording. For internal use only.
        """
        if self.model == "MRG":
            self.m_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.m_reclist, single_mod=False, key_node="_ref_m_axnode"
            )
            self.mp_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.mp_reclist, single_mod=False, key_node="_ref_mp_axnode"
            )
            self.h_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.h_reclist, single_mod=False, key_node="_ref_h_axnode"
            )
            self.s_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.s_reclist, single_mod=False, key_node="_ref_s_axnode"
            )

        else:  # should be Gaines, motor or sensory
            if self.model == "Gaines_motor":
                key_mod = "_motor"
            else:
                key_mod = "_sensory"

            self.m_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.m_reclist, single_mod=False, key_node="_ref_m_node" + key_mod
            )

            self.mp_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.mp_reclist, single_mod=False, key_node="_ref_mp_node" + key_mod
            )

            self.h_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.h_reclist, single_mod=False, key_node="_ref_h_node" + key_mod
            )

            self.s_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.s_reclist,
                single_mod=False,
                key_node="_ref_s_node" + key_mod,
                key_MYSA="_ref_s_mysa" + key_mod,
                key_FLUT="_ref_s_flut" + key_mod,
                key_STIN="_ref_s_stin" + key_mod,
            )

            self.n_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.n_reclist,
                single_mod=False,
                key_node="_ref_n_node" + key_mod,
                key_MYSA="_ref_n_mysa" + key_mod,
                key_FLUT="_ref_n_flut" + key_mod,
                key_STIN="_ref_n_stin" + key_mod,
            )

            self.q_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                self.q_reclist,
                single_mod=False,
                key_MYSA="_ref_q_mysa" + key_mod,
                key_FLUT="_ref_q_flut" + key_mod,
                key_STIN="_ref_q_stin" + key_mod,
            )

    def get_particles_values(self):
        """
        get the particules values at the end of simulation. For internal use only.
        """
        results = []
        results += [self.__get_recorders_from_list(self.m_reclist)]
        results += [self.__get_recorders_from_list(self.mp_reclist)]
        results += [self.__get_recorders_from_list(self.h_reclist)]
        results += [self.__get_recorders_from_list(self.s_reclist)]
        if self.model == "Gaines_motor" or self.model == "Gaines_sensory":
            results += [self.__get_recorders_from_list(self.n_reclist)]
            results += [self.__get_recorders_from_list(self.q_reclist)]
        return results

    def set_conductance_recorders(self):
        """
        Prepare the ionic channels conductance recording. For internal use only.
        """
        self.gna_reclist = neuron.h.List()
        self.gnap_reclist = neuron.h.List()
        self.gk_reclist = neuron.h.List()
        self.gl_reclist = neuron.h.List()
        if self.model == "MRG":
            self.__set_recorders_with_key(
                self.gna_reclist, single_mod=False, key_node="_ref_gna_axnode"
            )
            self.__set_recorders_with_key(
                self.gnap_reclist, single_mod=False, key_node="_ref_gnap_axnode"
            )
            self.__set_recorders_with_key(
                self.gk_reclist, single_mod=False, key_node="_ref_gk_axnode"
            )
            self.__set_recorders_with_key(
                self.gl_reclist, single_mod=False, key_node="_ref_gl_axnode"
            )
            if not self.rec == "nodes":
                self.gi_reclist = neuron.h.List()
                self.__set_recorders_with_key(
                    self.gi_reclist,
                    single_mod=False,
                    key_MYSA="_ref_g_pas",
                    key_FLUT="_ref_g_pas",
                    key_STIN="_ref_g_pas",
                )
        else:  # should be Gaines motor or sensory
            self.gkf_reclist = neuron.h.List()
            self.gq_reclist = neuron.h.List()
            if self.model == "Gaines_motor":
                key_mod = "_motor"
            else:
                key_mod = "_sensory"
            self.__set_recorders_with_key(
                self.gna_reclist,
                single_mod=False,
                key_node="_ref_gna_node" + key_mod,
            )
            self.__set_recorders_with_key(
                self.gnap_reclist,
                single_mod=False,
                key_node="_ref_gnap_node" + key_mod,
            )
            self.__set_recorders_with_key(
                self.gk_reclist,
                single_mod=False,
                key_node="_ref_gk_node" + key_mod,
                key_MYSA="_ref_gk_mysa" + key_mod,
                key_FLUT="_ref_gk_flut" + key_mod,
                key_STIN="_ref_gk_stin" + key_mod,
            )
            self.__set_recorders_with_key(
                self.gl_reclist,
                single_mod=False,
                key_node="_ref_gl_node" + key_mod,
                key_MYSA="_ref_gl_mysa" + key_mod,
                key_FLUT="_ref_gl_flut" + key_mod,
                key_STIN="_ref_gl_stin" + key_mod,
            )
            self.__set_recorders_with_key(
                self.gkf_reclist,
                single_mod=False,
                key_node="_ref_gkf_node" + key_mod,
                key_MYSA="_ref_gkf_mysa" + key_mod,
                key_FLUT="_ref_gkf_flut" + key_mod,
                key_STIN="_ref_gkf_stin" + key_mod,
            )
            if not self.rec == "nodes":
                self.__set_recorders_with_key(
                    self.gq_reclist,
                    single_mod=False,
                    key_MYSA="_ref_gq_mysa" + key_mod,
                    key_FLUT="_ref_gq_flut" + key_mod,
                    key_STIN="_ref_gq_stin" + key_mod,
                )

    def get_ionic_conductance(self):
        """
        get the ionic channels conductance at the end of simulation. For internal use only.
        """
        results = []
        results += [self.__get_recorders_from_list(self.gna_reclist)]
        results += [self.__get_recorders_from_list(self.gnap_reclist)]
        results += [self.__get_recorders_from_list(self.gk_reclist)]
        results += [self.__get_recorders_from_list(self.gl_reclist)]
        if self.model == "MRG":
            if self.rec == "nodes":
                results += [np.zeros((self.axonnodes, self.t_len))]
            else:
                results += [self.__get_recorders_from_list(self.gi_reclist)]
        else:  # should be Gaines, motor or sensory
            results += [self.__get_recorders_from_list(self.gkf_reclist)]
            if self.rec == "nodes":
                results += [np.zeros((self.axonnodes, self.t_len))]
            else:
                results += [self.__get_recorders_from_list(self.gq_reclist)]
        return results

    def get_membrane_conductance(self):
        """
        get the total membrane conductance at the end of simulation. For internal use only.
        """
        return sum(self.get_ionic_conductance())

    def get_membrane_capacitance(self):
        """
        get the membrane capacitance
        NB: [uF/cm^{2}] (see Neuron unit)
        """
        return self.__get_var_from_mod("_ref_cm")

    def get_myeline_conductance(self):
        """
        get the membrane capacitance
        NB: [S/cm^{2}] (see Neuron unit)
        """
        return self.__get_var_from_mod("_ref_xg")

    def get_myeline_capacitance(self):
        """
        get the membrane capacitance
        NB: [uF/cm^{2}] (see Neuron unit)
        """
        return self.__get_var_from_mod("_ref_xc")

    def set_Nav_recorders(self):
        """
        Prepare the markov model recording. For internal use only.
        """
        # Nav1.1 variables
        self.I_nav11_reclist = neuron.h.List()
        self.C1_nav11_reclist = neuron.h.List()
        self.C2_nav11_reclist = neuron.h.List()
        self.O1_nav11_reclist = neuron.h.List()
        self.O2_nav11_reclist = neuron.h.List()
        self.I1_nav11_reclist = neuron.h.List()
        self.I2_nav11_reclist = neuron.h.List()
        # Nav1.6 variables
        self.I_nav16_reclist = neuron.h.List()
        self.C1_nav16_reclist = neuron.h.List()
        self.C2_nav16_reclist = neuron.h.List()
        self.O1_nav16_reclist = neuron.h.List()
        self.O2_nav16_reclist = neuron.h.List()
        self.I1_nav16_reclist = neuron.h.List()
        self.I2_nav16_reclist = neuron.h.List()
        for k in self.Markov_Nav_modeled_NoR:
            # Nav1.1
            I_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_ina_na11a, sec=self.node[k]
            )
            C1_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_C1_na11a, sec=self.node[k]
            )
            C2_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_C2_na11a, sec=self.node[k]
            )
            O1_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_O1_na11a, sec=self.node[k]
            )
            O2_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_O2_na11a, sec=self.node[k]
            )
            I1_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_I1_na11a, sec=self.node[k]
            )
            I2_nav11 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_I2_na11a, sec=self.node[k]
            )
            self.I_nav11_reclist.append(I_nav11)
            self.C1_nav11_reclist.append(C1_nav11)
            self.C2_nav11_reclist.append(C2_nav11)
            self.O1_nav11_reclist.append(O1_nav11)
            self.O2_nav11_reclist.append(O2_nav11)
            self.I1_nav11_reclist.append(I1_nav11)
            self.I2_nav11_reclist.append(I2_nav11)
            # Nav1.6
            I_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_ina_na16a, sec=self.node[k]
            )
            C1_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_C1_na16a, sec=self.node[k]
            )
            C2_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_C2_na16a, sec=self.node[k]
            )
            O1_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_O1_na16a, sec=self.node[k]
            )
            O2_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_O2_na16a, sec=self.node[k]
            )
            I1_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_I1_na16a, sec=self.node[k]
            )
            I2_nav16 = neuron.h.Vector().record(
                self.node[k](0.5)._ref_I2_na16a, sec=self.node[k]
            )
            self.I_nav16_reclist.append(I_nav16)
            self.C1_nav16_reclist.append(C1_nav16)
            self.C2_nav16_reclist.append(C2_nav16)
            self.O1_nav16_reclist.append(O1_nav16)
            self.O2_nav16_reclist.append(O2_nav16)
            self.I1_nav16_reclist.append(I1_nav16)
            self.I2_nav16_reclist.append(I2_nav16)

    def get_Nav_values(self):
        """
        get the markov model at the end of simulation. For internal use only.
        """
        # Nav1.1 values
        I_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        C1_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        C2_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        O1_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        O2_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        I1_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        I2_nav11_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        for k in range(len(self.I_nav11_reclist)):
            I_nav11_ax[k, :] = np.asarray(self.I_nav11_reclist[k])
            C1_nav11_ax[k, :] = np.asarray(self.C1_nav11_reclist[k])
            C2_nav11_ax[k, :] = np.asarray(self.C2_nav11_reclist[k])
            O1_nav11_ax[k, :] = np.asarray(self.O1_nav11_reclist[k])
            O2_nav11_ax[k, :] = np.asarray(self.O2_nav11_reclist[k])
            I1_nav11_ax[k, :] = np.asarray(self.I1_nav11_reclist[k])
            I2_nav11_ax[k, :] = np.asarray(self.I2_nav11_reclist[k])
        # Nav1.6 values
        I_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        C1_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        C2_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        O1_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        O2_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        I1_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        I2_nav16_ax = np.zeros((len(self.Markov_Nav_modeled_NoR), self.t_len))
        for k in range(len(self.I_nav11_reclist)):
            I_nav16_ax[k, :] = np.asarray(self.I_nav16_reclist[k])
            C1_nav16_ax[k, :] = np.asarray(self.C1_nav16_reclist[k])
            C2_nav16_ax[k, :] = np.asarray(self.C2_nav16_reclist[k])
            O1_nav16_ax[k, :] = np.asarray(self.O1_nav16_reclist[k])
            O2_nav16_ax[k, :] = np.asarray(self.O2_nav16_reclist[k])
            I1_nav16_ax[k, :] = np.asarray(self.I1_nav16_reclist[k])
            I2_nav16_ax[k, :] = np.asarray(self.I2_nav16_reclist[k])
        return (
            I_nav11_ax,
            C1_nav11_ax,
            C2_nav11_ax,
            O1_nav11_ax,
            O2_nav11_ax,
            I1_nav11_ax,
            I2_nav11_ax,
            I_nav16_ax,
            C1_nav16_ax,
            C2_nav16_ax,
            O1_nav16_ax,
            O2_nav16_ax,
            I1_nav16_ax,
            I2_nav16_ax,
        )



nrv.rasterize(axon_sim, "V_mem")


list_keys = [
    "ID",
    "L",
    "V_mem_raster_position",
    "V_mem_raster_x_position",
    "V_mem_raster_time_index",
    "V_mem_raster_time",
    "myelinateds",
    "intra_stim_starts",
    "intra_stim_positions",
]

removable_keys = []
for key in axon_sim:
    if not key in list_keys:
        removable_keys += [key]

for key in removable_keys:
    nrv.remove_key(axon_sim, key, verbose=self.verbose)



"""
NRV-fascicule generator
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import math
import os

import matplotlib.pyplot as plt
import numba
import numpy as np
from scipy import spatial, stats
from scipy.optimize import curve_fit
from scipy.stats import rv_continuous

from ..backend.log_interface import pass_info, progression_popup

# WARNING:
# no prompt message for numpy division by zeros: handled in the code !!!
np.seterr(divide="ignore", invalid="ignore")

# verbosity level
fg_verbose = True

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()

# get the built-in material librairy
dir_path = os.environ["NRVPATH"] + "/_misc"
stat_library = os.listdir(dir_path + "/stats/")

#################################################
## Nerve composition statistics and generators ##
#################################################
myelinated_stats = [
    "Schellens_1",
    "Schellens_2",
    "Ochoa_M",
    "Jacobs_9_A",
    "Jacobs_9_B",
    "Jacobs_9_C",
    "Jacobs_9_D",
]
unmyelinated_stats = [
    "Ochoa_U",
    "Jacobs_11_A",
    "Jacobs_11_B",
    "Jacobs_11_C",
    "Jacobs_11_D",
]


def load_stat(stat_name):
    """
    Load a statistic stored in the included librairy or specified by the user in a csv file (first corlumn: axon diameter bin start - second corlumn: proportion)

    Parameters
    ----------
    stat_name : str
        name of the statistic in the librairy or path to a new librairy in csv

    Returns
    -------
    diameters   : list
        diameters as start of bins of the histogram
    presence    : list
        quantities of presence of each bin in the histogram
    """
    f_in_librairy = str(stat_name) + ".csv"
    if f_in_librairy in stat_library:
        stat_file = np.genfromtxt(dir_path + "/stats/" + f_in_librairy, delimiter=",")
    else:
        # stat_file = np.genfromtxt(dir_path + "/stats/" + f_in_librairy, delimiter=",")
        stat_file = np.genfromtxt(f_in_librairy, delimiter=",")
    diameters = stat_file[:, 0]
    presence = stat_file[:, 1]
    return diameters, presence


def one_Gamma(x, a1, beta1, c1):
    """
    Gamma function, mono-modal, to interpolate unmyelinated statistics.

    Parameters
    ----------
    x       : float
        variable of the Gamma function, will correspond to diameter
    a1      : float
        shape parameter
    beta1   : float
        scale parameter
    c1      : float
        gain applied to the gamma function

    Note
    ----
    location of the gamma pdf function is fixed to 0.2, diameters will be interpolated above this minimal value.
    """
    return c1 * (stats.gamma.pdf(x, a1, scale=1 / beta1, loc=0.2))


def two_Gamma(x, a1, beta1, c1, a2, beta2, c2, transition):
    """
    Gamma function, bi-modal, to interpolate myelinated statistics.

    Parameters
    ----------
    x           : float
        variable of the Gamma function, will correspond to diameter
    a1          : float
        shape parameter on the first (lowest) lobe
    beta1       : float
        scale parameter on the first (lowest) lobe
    c1          : float
        gain applied to the gamma function on the first (lowest) lobe
    a2          : float
        shape parameter on the second (highest) lobe
    beta2       : float
        scale parameter on the second (highest) lobe
    c2          : float
        gain applied to the gamma function on the second (highest) lobe
    transition  :float
        diameter value at the transition between the two lobes

    Note
    ----
    location of the gamma pdf first function is fixed to 2, diameters will be interpolated above this minimal value. This corresponds to the minimal A-delta diamter tolerated value.
    """
    return c1 * (stats.gamma.pdf(x, a1, scale=1 / beta1, loc=2)) + c2 * (
        stats.gamma.pdf(x, a2, scale=1 / beta2, loc=transition)
    )


class nerve_gen_one_gamma(rv_continuous):
    """
    Class to create specific repartition law generator for unmyelinated axons. Inherits from scipy rv_continuous class. Some methods are overwritten
    """

    def set_bounds(self, v_min, v_max):
        """
        Set bounds to the generator.

        Parameters
        ----------
        v_min   : float
            minimum value
        v_max   : float
            maximum value
        """
        self.a = v_min
        self.b = v_max

    def configure(self, a1, beta1, c1):
        """
        Set the value from interpolated data

        Parameters
        ----------
        a1      : float
            shape parameter
        beta1   : float
            scale parameter
        c1      : float
            gain applied to the gamma function

        """
        self.a1 = a1
        self.beta1 = beta1
        self.c1 = c1

    def _pdf(self, x):
        """
        overwritting the pdf to custom law, same definition as one_Gamma described upper.
        """
        return self.c1 * (stats.gamma.pdf(x, self.a1, scale=1 / self.beta1, loc=0.2))


class nerve_gen_two_gamma(rv_continuous):
    """
    Class to create specific repartition law generator for myelinated axons. Inherits from scipy rv_continuous class. Some methods are overwritten
    """

    def set_bounds(self, v_min, v_max):
        """
        Set bounds to the generator.

        Parameters
        ----------
        v_min   : float
            minimum value
        v_max   : float
            maximum value
        """
        self.a = v_min
        self.b = v_max

    def configure(self, a1, beta1, c1, a2, beta2, c2, transition):
        """
        Set the value from interpolated data

        Parameters
        ----------
        a1          : float
            shape parameter on the first (lowest) lobe
        beta1       : float
            scale parameter on the first (lowest) lobe
        c1          : float
            gain applied to the gamma function on the first (lowest) lobe
        a2          : float
            shape parameter on the second (highest) lobe
        beta2       : float
            scale parameter on the second (highest) lobe
        c2          : float
            gain applied to the gamma function on the second (highest) lobe
        transition  :float
            diameter value at the transition between the two lobes

        """
        self.a1 = a1
        self.a2 = a2
        self.beta1 = beta1
        self.beta2 = beta2
        self.c1 = c1
        self.c2 = c2
        self.transition = transition

    def _pdf(self, x):
        """
        overwritting the pdf to custom law, same definition as two_Gamma described upper.
        """
        return self.c1 * (
            stats.gamma.pdf(x, self.a1, scale=1 / self.beta1, loc=2)
        ) + self.c2 * (
            stats.gamma.pdf(x, self.a2, scale=1 / self.beta2, loc=self.transition)
        )


def create_generator_from_stat(stat, myelinated=True, dmin=None, dmax=None):
    """
    Create a statistical generator (type rv_continuous) for a given statistic.

    Parameters
    ----------
    stat        : str
        name of the statistic in the librairy or path to a new librairy in csv
    myelinated  : bool
        True if the statistic is describing myelinated diameters, esle False. If the statisticis in the librairy, this will be automatically chosen.
    dmin        : float
        minimal diameter to consider, in um. If None, the minimal value of the statistic is taken
    dmax        : float
        minimal diameter to consider, in um. If None, the maximal value of the statistic plus bin size is taken

    Returns
    -------
    generator   : rv_continuous
        Generator object corresponding to the statistic.
    """
    # Check if myelinated or not for standard stats
    if stat in myelinated_stats:
        myelinated = True
    elif stat in unmyelinated_stats:
        myelinated = False
    # load the stat
    diameters, presence = load_stat(stat)
    # chose min max values for the axon diameters
    if dmin is None:
        dmin = min(diameters)
    if dmax is None:
        bin_size = diameters[1] - diameters[0]
        dmax = max(diameters) + bin_size
    # perform stat fitting and create generator
    if myelinated:
        popt1, pcov1 = curve_fit(
            two_Gamma,
            xdata=diameters,
            ydata=presence,
            bounds=(
                [1, 2, 0, 1, 0, 0, 2],
                [np.inf, 15, np.inf, np.inf, 15, np.inf, 14],
            ),
        )
        generator = nerve_gen_two_gamma()
        generator.set_bounds(dmin, dmax)
        generator.configure(*popt1)
    else:
        popt1, pcov1 = curve_fit(
            one_Gamma,
            xdata=diameters,
            ydata=presence,
            bounds=([1, 0, 0], [np.inf, 10, np.inf]),
        )
        generator = nerve_gen_one_gamma()
        generator.set_bounds(dmin, dmax)
        generator.configure(*popt1)
    return generator, popt1, pcov1


def create_axon_population(
    N, percent_unmyel=0.7, M_stat="Schellens_1", U_stat="Ochoa_U"
):
    """
    Create a virtual population of axons (no Neuron implementation, axon class not called) of a controled number, with controlled statistics.

    Parameters
    ----------
    N               : int
        Number of axon to generate for the population (Unmyelinated and myelinated)
    percent_unmyel  : float
        ratio of unmyelinated axons in the population. Should be between 0 and 1.
    M_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
    U_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition

    Returns
    -------
    axons_diameters     : np.array
        Array of length N, containing all the diameters of the generated axon population
    axons_type          : np.array
        Array of length N, containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    # create generators
    M_gen, M_popt, M_pcov = create_generator_from_stat(M_stat)
    U_gen, U_popt, U_pcov = create_generator_from_stat(U_stat)
    # number of myelinated and unmyelinated axons to create
    U = int(N * percent_unmyel)
    M = N - U
    pass_info(
        "On "
        + str(N)
        + " axons to generate, there are "
        + str(M)
        + " Myelinated and "
        + str(U)
        + " Unmyelinated"
    )
    # generate the myelinated axons
    xspace1 = np.linspace(1, 20, num=500)
    if len(M_popt) < 4:
        data = one_Gamma(xspace1, *M_popt)
    else:
        data = two_Gamma(xspace1, *M_popt)
    data = data / np.sum(data)
    M_diam_list = np.random.choice(xspace1, M, p=data)
    M_type = np.ones(M)

    # generate the unmyelinated axons
    xspace1 = np.linspace(0.1, 3, num=500)
    data = one_Gamma(xspace1, *U_popt)
    data = data / np.sum(data)
    U_diam_list = np.random.choice(xspace1, U, p=data)
    U_type = np.zeros(U)

    shuffle_mask = np.random.permutation(N)
    axons_diameters = np.concatenate((M_diam_list, U_diam_list))
    axons_type = np.concatenate((M_type, U_type))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type, M_diam_list, U_diam_list


def fill_area_with_axons(
    A, percent_unmyel=0.7, FVF=0.55, M_stat="Schellens_1", U_stat="Ochoa_U"
):
    """
    Create a virtual population of axons (no Neuron implementation, axon class not called) to fill a specified area, with controlled statistics.

    Parameters
    ----------
    A               : float
        surface area to fill, in um**2
    percent_unmyel  : float
        ratio of unmyelinated axons in the population. Should be between 0 and 1.
    FVF             :
        Fiber Volume Fraction estimated for the area. By default set to 0.55
    M_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
    U_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition

    Returns
    -------
    axons_diameters     : np.array
        Array  containing all the diameters of the generated axon population
    axons_type          : np.array
        Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    # create generators
    M_gen, M_popt, M_pcov = create_generator_from_stat(M_stat)
    M_diam_list = []
    M_max_diam_value = M_gen.b - 0.01 * (M_gen.b - M_gen.a)
    U_gen, U_popt, U_pcov = create_generator_from_stat(U_stat)
    U_diam_list = []
    U_max_diam_value = U_gen.b - 0.01 * (U_gen.b - U_gen.a)
    # area
    A_ax = 0
    A_total = 0
    # axon lists
    M_diam_list = []
    U_diam_list = []
    # loop
    while A_total < A:
        filled = A_total / A
        progression_popup(
            round(filled * 100), 100, begin_message="\t Area filled at ", endl="\r"
        )
        # print('\t Area filled at ' + f"{round(filled*100)}" + ' percent', end="\r")
        U_or_M = np.random.uniform(0, 1)
        if U_or_M < percent_unmyel:
            # generate a unmyelinated axon
            prov_diam = U_gen.rvs(1, size=1)[0]
            while prov_diam > U_max_diam_value:
                prov_diam = U_gen.rvs(1, size=1)[0]
            U_diam_list.append(prov_diam)
            A_ax += np.power(U_diam_list[-1] / 2, 2) * np.pi
        else:
            prov_diam = M_gen.rvs(1, size=1)[0]
            while prov_diam > M_max_diam_value:
                prov_diam = M_gen.rvs(1, size=1)[0]
            M_diam_list.append(prov_diam)
            A_ax += np.power(M_diam_list[-1] / 2, 2) * np.pi
        A_total = A_ax / FVF
    # concatenate and shuffle
    M = len(M_diam_list)
    U = len(U_diam_list)
    N = M + U
    M_type = np.ones(M)
    U_type = np.zeros(U)
    shuffle_mask = np.random.permutation(N)
    axons_diameters = np.concatenate((M_diam_list, U_diam_list))
    axons_type = np.concatenate((M_type, U_type))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type, M_diam_list, U_diam_list


def shuffle_population(axons_diameters, axons_type):
    """
    Shuffle an axonal population

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)

    Returns
    -------
    axons_diameters : np.array
        shuffled axons diamters
    axons_type      : np.array
        corrresponding axons type
    """
    shuffle_mask = np.random.permutation(len(axons_diameters))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type


def save_axon_population(f_name, axons_diameters, axons_type, comment=None):
    """
    Save an axonal population to a file

    Parameters
    ----------
    f_name          : str
        name of the file to store the population
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)
    comment         : str
        comment added in the header of the file, optional
    """
    f = open(f_name, "w")
    if comment is not None:
        line = "# " + comment
        f.write(line)
    for k in range(len(axons_diameters)):
        line = str(axons_diameters[k]) + ", " + str(axons_type[k]) + "\n"
        f.write(line)
    f.close()


def load_axon_population(f_name):
    """
    Load a population from a file

    Parameters
    ----------
    f_name          : str
        name of the file where the population is stored

    Returns
    -------
    axons_diameters     : np.array
        Array containing all the diameters of the generated axon population
    axons_type          : np.array
        Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    population_file = np.genfromtxt(f_name, delimiter=",", comments="#")
    axons_diameters = population_file[:, 0]
    axons_type = population_file[:, 1]
    ind_myel = np.argwhere(axons_type == 1)
    ind_unmyel = np.argwhere(axons_type == 0)
    M_diam_list = axons_diameters[ind_myel]
    U_diam_list = axons_diameters[ind_unmyel]
    return axons_diameters, axons_type, M_diam_list, U_diam_list


#############################################
#############################################
#############################################
def axon_packer(
    diameters,
    Delta=0,
    y_gc=0,
    z_gc=0,
    max_iter=20000,
    probe=100,
    monitor=False,
    monitoring_Folder="",
    monitoring_Niter=100,
    v_att=0.01,
    v_rep=0.1,
):
    """
    Axon Packing algorithm: this operation takes a vector of diameter (random population) and places it at best. The used algorithm is largely based on [1]

    Parameters
    ----------
    diameters           : np.array
        Array containing all the diameters of the axon population to pack
    Delta               : float
        minimal inter-axon distance to respect before considering collision, in um
    y_gc                : float
        y coordinate of the gravity center for the packing, in um
    z_gc                : float
        z coordinate of the gravity center for the packing, in um
    max_iter            : int
        Max. number of iterations
    probe               : int
        Number of iterations between two Fiber Volume Fraction probing
    monitor             : bool
        if True, the packing process will be monitored and some steps are plotted and saved
    monitoring_Folder   : str
        in case of monitoring, folder where the images of the packing process are stored
    monitoring_Niter    : str
        in case of monitoring, number of iterations between two monitoring images
    v_att               : float
        vector norm for attraction velocity, in um per iteration
    v_rep               : float
        vector norm for repulsion velocity, in um per iteration

    Returns
    -------
    y_axons         : np.array
        Array containing the y coordinates of axons, in um
    z_axons         : np.array
        Array containing the z coordinates of axons, in um
    iteration       : int
        number of iterations performed
    FVF             : np.array
        probed values for the Fiber Volume Fraction
    probed_iter     : np.array
        index of the FVF probed iterations

    Note
    ----
    - scientific reference
        [1] Mingasson, T., Duval, T., Stikov, N., and Cohen-Adad, J. (2017). AxonPacking: an open-source software to simulate arrangements of axons in white matter. Frontiers in neuroinformatics, 11, 5.
    - dev Note
        the algorithm perform a fixed number of iteration, the code could evoluate to tke into account the FVF convergence, however this value depends on the range on number of axons.
    """
    N = len(diameters)
    max_diam = max(diameters)
    ##########
    ## INIT ##
    ##########
    ### create an initial square grid with the population of axons
    N_init = int(math.ceil(np.sqrt(N)) ** 2)
    N_side = int(np.sqrt(N_init))
    size_init = np.sqrt(N_init * (max_diam + Delta) ** 2)
    placement_vector = np.linspace(
        -((size_init / 2) - (size_init / (2 * N_side))),
        ((size_init / 2) - (size_init / (2 * N_side))),
        num=N_side,
        endpoint=True,
    )
    y_axons = np.tile(placement_vector, N_side) + y_gc
    z_axons = np.repeat(placement_vector, N_side) + z_gc
    ### remove unwanted places
    N_remove = N_init - N
    ind_to_delete = np.random.choice(len(y_axons), size=N_remove, replace=False)
    y_axons = np.delete(y_axons, ind_to_delete)
    z_axons = np.delete(z_axons, ind_to_delete)
    ### compute total total fiber surface and evaluate surface for FVF computation
    ax_areas = np.power(diameters / 2, 2) * np.pi
    A_axons = np.sum(ax_areas)
    size_monitor_square = np.sqrt(A_axons * 1.4)
    monitor_area = A_axons * 1.4
    FVF = []
    probed_iter = []
    if fg_verbose:
        pass_info("Axon Packer: initial grid computed...")
    if monitor:
        plot_situation(
            diameters, y_axons, z_axons, size_init, title="Init", y_gc=y_gc, z_gc=z_gc
        )
        f_name = monitoring_Folder + "Packer_0.png"
        plt.savefig(f_name)
        plt.close()
    ################
    ## ITERATIONS ##
    ################
    if fg_verbose:
        pass_info("Axon Packer: Begining iterations")
    iteration = 0
    while iteration < max_iter:
        iteration += 1
        if fg_verbose:
            progression_popup(
                iteration - 1, max_iter, begin_message="\t iteration ", endl="\r"
            )
            # print('\t iteration ' + f"{iteration}" + '/' + str(max_iter), end="\r")
        # compute P matrix
        P = compute_P_matrix(diameters, y_axons, z_axons, Delta, N)
        # P = compute_P_fast(diameters, y_axons, z_axons, Delta)
        # check overlap
        colapse = np.argwhere(P < 0)
        all_colapsed_ind = np.unique(colapse)
        # compute velocity, initialy consider that all velocities are attraction
        v_y, v_z = compute_attraction_velocities(y_axons, z_axons, y_gc, z_gc, N, v_att)
        #### prevent from division by 0 in the upper line
        np.nan_to_num(v_y, copy=False, nan=0.0)
        np.nan_to_num(v_z, copy=False, nan=0.0)
        v_y[all_colapsed_ind] = 0  # neutralize velocity where should be repulsion
        v_z[all_colapsed_ind] = 0  # same
        # compute velocities for colapsing cases
        v_y, v_z = handle_collisions(
            y_axons, z_axons, v_y, v_z, all_colapsed_ind, colapse, v_rep
        )
        # compute new positions
        y_axons, z_axons = update_positions(y_axons, z_axons, v_y, v_z)
        # compute FVF
        if iteration % probe == 0:
            # removing axons too much to the left
            ax_left_pts = y_axons - diameters / 2
            FVF_mask = np.argwhere(ax_left_pts > (-size_monitor_square / 2 + y_gc))
            # removing axons too much to the right
            ax_right_pts = y_axons + diameters / 2
            FVF_mask = np.intersect1d(
                FVF_mask, np.argwhere(ax_right_pts < (size_monitor_square / 2 + y_gc))
            )
            # removing the axons that are too high
            ax_up_pts = z_axons + diameters / 2
            FVF_mask = np.intersect1d(
                FVF_mask, np.argwhere(ax_up_pts < (size_monitor_square / 2 + z_gc))
            )
            # removing the axons that are too low
            ax_down_pts = z_axons - diameters / 2
            FVF_mask = np.intersect1d(
                FVF_mask, np.argwhere(ax_down_pts > (-size_monitor_square / 2 + z_gc))
            )
            FVF.append(np.sum(ax_areas[FVF_mask]) / monitor_area)
            probed_iter.append(iteration)
        # monitoring
        if (iteration) % monitoring_Niter == 0 and monitor:
            plot_situation(
                diameters,
                y_axons,
                z_axons,
                size_init,
                title=str(iteration) + " Iterations",
                y_gc=y_gc,
                z_gc=z_gc,
            )
            f_name = monitoring_Folder + "Packer_" + str(iteration) + ".png"
            plt.savefig(f_name)
            plt.close()
    pass_info("Axon Packer: Iterations done")
    if monitor:
        plot_situation(
            diameters,
            y_axons,
            z_axons,
            size_init,
            title=str(iteration) + " Iterations",
            y_gc=y_gc,
            z_gc=z_gc,
        )
        f_name = monitoring_Folder + "Packer_final.png"
        plt.savefig(f_name)
        plt.close()
    return y_axons, z_axons, iteration, np.asarray(FVF), np.asarray(probed_iter)


@numba.jit(fastmath=True, cache=True)
def compute_attraction_velocities(y_axons, z_axons, y_gc, z_gc, N, v_att):
    """
    Computes attraction velocity for axon packing. Just in time compiled to speed up. For internal use only.
    """
    dy = np.ones(N) * y_gc - y_axons
    dz = np.ones(N) * z_gc - z_axons
    dist = np.sqrt((y_axons - y_gc) ** 2 + (z_axons - z_gc) ** 2)
    v_y = v_att * dy / dist
    v_z = v_att * dz / dist  # same
    return v_y, v_z


@numba.njit(fastmath=True, cache=True)
def compute_P_matrix(diameters, y_axons, z_axons, Delta, N):
    """
    Computes the proximity matrix for axon packing. Just in time compiled to speed up. For internal use only.
    """
    P = np.zeros((N, N))
    for i in range(N):
        for j in range(i + 1, N):
            # P[i, j] = np.sqrt((y_axons[i] - y_axons[j])**2 + (z_axons[i] - z_axons[j])**2) -\
            # ((diameters[i] + diameters[j])/2 + Delta)
            # not hmogeneous to a distance, but work the same at the end and one square root less to compute...
            P[i, j] = (
                (y_axons[i] - y_axons[j]) ** 2 + (z_axons[i] - z_axons[j]) ** 2
            ) - ((diameters[i] + diameters[j]) / 2 + Delta) ** 2
    return P


def compute_P_fast(diameters, y_axons, z_axons, Delta):
    """
    Compyte the proximity matrix for axon packing. Based on numpy and scipy. Slower (just a bit) than JIT version on dev. For internal use only
    """
    points = np.column_stack((y_axons, z_axons))
    N = len(diameters)
    diams = np.tile(diameters, (N, 1))
    # diams = diameters.repeat(N).reshape((-1, N))
    P = spatial.distance.cdist(points, points, "euclidean") - (
        (diams + np.transpose(diams)) * (1 / 2) + np.ones((N, N)) * Delta
    )
    np.fill_diagonal(P, 0)
    return np.triu(P)


@numba.njit(fastmath=True, cache=True)
def handle_collisions(y_axons, z_axons, v_y, v_z, all_colapsed_ind, colapse, v_rep):
    """
    Handle collisions between axons for axon packing. Just in time compiled (doble loop...). For internal use only.
    """
    for k in all_colapsed_ind:
        sum_y = 0
        sum_z = 0
        for i in range(len(colapse)):
            duplet = colapse[i]
            if k == duplet[0]:
                sum_y += y_axons[k] - y_axons[duplet[1]]
                sum_z += z_axons[k] - z_axons[duplet[1]]
            elif k == duplet[1]:
                sum_y += y_axons[k] - y_axons[duplet[0]]
                sum_z += z_axons[k] - z_axons[duplet[0]]
        if sum_y == 0 and sum_z == 0:
            sum_y = 1e-12
            sum_z = 1e-12
        v_y[k] = v_rep * (sum_y / (np.sqrt(sum_y**2 + sum_z**2)))
        v_z[k] = v_rep * (sum_z / (np.sqrt(sum_y**2 + sum_z**2)))
    return v_y, v_z


@numba.njit(fastmath=True, cache=True)
def update_positions(y_axons, z_axons, v_y, v_z):
    """
    Compute new positions considering velicities for axon packing. Just in time compiled to speed up. For internal use only.
    """
    return y_axons + v_y, z_axons + v_z


def delete_collisions(axons_diameters, axons_type, y_axons, z_axons, Delta=0):
    """
    Delete collision cases in a placed population. In case of a detected collision, the axon or smaller diameter is removed.

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)
    y_axons         : np.array
        y coordinate of the axons to store, in um
    z_axons         : np.array
        z coordinate of the axons to store, in um
    Delta           : float
        space between axon fibers under which collision is considered, in um

    Returns
    -------
    new_axons_diameters : np.array
        array containing the axons diameters without collisions
    new_axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)
    new_y_axons     : np.array
        y coordinate of the axons to store, in um
    new_z_axons     : np.array
        z coordinate of the axons to store, in um
    """
    N = len(axons_diameters)
    P = compute_P_matrix(axons_diameters, y_axons, z_axons, Delta, N)
    colapse = np.argwhere(P < 0)
    flag_Collision = colapse.size != 0

    while flag_Collision:
        # checking which axons to delete
        ind_to_delete = []
        handled_collisions = []
        for collision in colapse:
            if (collision[0] not in handled_collisions) and (
                collision[1] not in handled_collisions
            ):
                handled_collisions.append(collision[0])
                handled_collisions.append(collision[1])
                if axons_diameters[collision[0]] < axons_diameters[collision[1]]:
                    ind_to_delete.append(collision[0])
                else:
                    ind_to_delete.append(collision[1])
        # deleting problematic axons
        mask = np.ones(N, dtype=bool)
        mask[ind_to_delete] = False
        axons_diameters = axons_diameters[mask]
        axons_type = axons_type[mask]
        y_axons = y_axons[mask]
        z_axons = z_axons[mask]
        N = len(axons_diameters)
        # check if there are remaing collisions
        P = compute_P_matrix(axons_diameters, y_axons, z_axons, Delta, N)
        colapse = np.argwhere(P < 0)
        flag_Collision = colapse.size != 0

    return axons_diameters, axons_type, y_axons, z_axons


def plot_situation(diameters, y_axons, z_axons, size, title=None, y_gc=0, z_gc=0):
    """
    Discplay a population of axons. Doesn't return but declares a matplotlib figure

    Parameters
    ----------
    diameters   : np.array
        diamters of the axons to display, in um
    y_axons     : np.array
        y coordinate of the axons to display, in um
    z_axons     : np.array
        z coordinate of the axons to display, in um
    size        : float
        size of the window as a square side, in um
    title       : str
        title of the figure
    y_gc        : float
        y coordinate of the gravity, in um
    z_gc        : float
        z coordinate of the gravity, in um
    """
    circles = []
    for k in range(len(diameters)):
        circles.append(
            plt.Circle(
                (y_axons[k], z_axons[k]), diameters[k] / 2, color="r", fill=False
            )
        )
    fig, ax = plt.subplots(figsize=(8, 8))
    for circle in circles:
        ax.add_patch(circle)
    if title is not None:
        plt.title(title)
    plt.xlim(-size / 2 + y_gc, size / 2 + y_gc)
    plt.ylim(-size / 2 + z_gc, size / 2 + z_gc)


def save_placed_axon_population(
    f_name, axons_diameters, axons_type, y_axons, z_axons, comment=None
):
    """
    Save a placed axonal population to a file

    Parameters
    ----------
    f_name          : str
        name of the file to store the population
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    comment         : str
        comment added in the header of the file, optional
    """
    f = open(f_name, "w")
    if comment is not None:
        line = "# " + comment
        f.write(line)
    for k in range(len(axons_diameters)):
        line = (
            str(axons_diameters[k])
            + ", "
            + str(axons_type[k])
            + ", "
            + str(y_axons[k])
            + ", "
            + str(z_axons[k])
            + "\n"
        )
        f.write(line)
    f.close()


def load_placed_axon_population(f_name):
    """
    Load a placed population from a file

    Parameters
    ----------
    f_name          : str
        name of the file where the population is stored

    Returns
    -------
    axons_diameters     : np.array
        Array containing all the diameters of the generated axon population
    axons_type          : np.array
        Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    y_axons     : np.array
        y coordinate of the axons, in um
    z_axons     : np.array
        z coordinate of the axons, in um
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    population_file = np.genfromtxt(f_name, delimiter=",", comments="#")
    axons_diameters = population_file[:, 0]
    axons_type = population_file[:, 1]
    y_axons = population_file[:, 2]
    z_axons = population_file[:, 3]
    ind_myel = np.argwhere(axons_type == 1)
    ind_unmyel = np.argwhere(axons_type == 0)
    M_diam_list = axons_diameters[ind_myel]
    U_diam_list = axons_diameters[ind_unmyel]
    return axons_diameters, axons_type, y_axons, z_axons, M_diam_list, U_diam_list



from abc import ABCMeta, abstractmethod
import os
from typing import Any
import numpy as np
import matplotlib.pyplot as plt
from time import perf_counter, asctime, localtime
from multiprocessing import cpu_count
import traceback
import json
import sys
import pyswarms as ps
from pyswarms.backend.topology import Star, Ring
import scipy.optimize as scpopt


from ..backend.NRV_Class import NRV_class
from ..backend.parameters import parameters
from ..backend.log_interface import rep_nrv, rise_error
from .optim_utils.optim_results import optim_results

dir_path = os.environ["NRVPATH"] + "/_misc"
dir_path + "/log/NRV.log"
# mymodule.py

class Optimizer(NRV_class, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, method=None):
        super().__init__()
        self._method = method
        self.swarm_optimizer = False
    
    def minimize(self, f, **kwargs):
        self.set_parameters(**kwargs)
        results = optim_results(self.save(save=False))
        results["date"] = asctime(localtime())
        results["status"] = "Processing"
        return results

    def __call__(self, f,**kwargs: Any) -> optim_results:
        return self.minimize(f,**kwargs)


class scipy_optimizer(Optimizer):
    def __init__(
        self,
        method=None,
        x0=None,
        args=(),
        jac=None,
        hess=None,
        hessp=None,
        bounds=None,
        constraints=(),
        tol=None,
        callback=None,
        maxiter = None,
        options=None,
        dimension=None,
        normalize=False,
    ):
        if method is None:
            super().__init__("scipy_default")
        else:
            super().__init__("scipy_"+method)
        self.x0 = x0
        self.args = args
        self.scipy_method = method
        self.jac = jac
        self.hess = hess
        self.hessp = hessp
        self.bounds = bounds
        self.constraints = constraints
        self.tol = tol
        self.callback = callback
        self.options = options or {}
        self.maxiter = maxiter
        self.dimensions = dimension
        self.normalize = normalize
        self.scale_translation = None
        self.scale_homothety = None
        self.scaled_bounds = None

    def __update_dimensions(self, results):
        """
        
        """
        if self.bounds is None:
            if self.dimensions is not None:
                self.bounds = [(None, None) for _ in range(self.dimensions)]
        if self.x0 is None:
            if self.bounds is None:
                rise_error("scipy optimizer: at least x0 or boundaries should be initiate")
            else:
                self.dimensions = len(self.bounds)
                self.x0 = []
                
                for i in range(self.dimensions):
                    min, max = self.bounds[i]
                    if min is None:
                        min = np.finfo(np.float32).min
                    if max is None:
                        max = np.finfo(np.float32).max
                    self.x0 += [np.random.uniform(min, max)]
                results["x0"] = self.x0
        else:
            self.dimensions = len(self.x0)
            results["dimensions"] = self.dimensions

    def __normalize_bound(self, results):
        self.scale_translation = np.zeros(self.dimensions)
        self.scale_homothety =  np.ones(self.dimensions)
        if not self.normalize:
            self.scaled_bounds = self.bounds
        else:
            self.scaled_bounds = []
            for i, bd in enumerate(self.bounds):
                self.scaled_bounds += [(0,1)]
                self.scale_translation[i] = bd[0]
                self.scale_homothety[i] = bd[1] - bd[0]
        results['scale_translation'] = self.scale_translation
        results['scale_homothety'] = self.scale_homothety
        results['scaled_bounds'] = self.scaled_bounds

    def minimize(self, f, **kwargs):
        results = super().minimize(f, **kwargs)
        if self.maxiter is not None: 
            self.options["maxiter"] = self.maxiter
        self.__update_dimensions(results)
        self.__normalize_bound(results)
        results["cost_history"] = []
        results["position_history"] = []

        def f_history(x):
            x = (results['scale_homothety'] * x) + results['scale_translation']
            c = f(x)
            results["position_history"].append([x])
            results["cost_history"].append(c)
            return c

        t0 = perf_counter()
        res = scpopt.minimize(
            fun=f_history,
            x0= self.x0,
            args= self.args,
            method= self.scipy_method,
            jac= self.jac,
            hess= self.hess,
            hessp= self.hessp,
            bounds= self.scaled_bounds,
            constraints= self.constraints,
            tol= self.tol,
            callback= self.callback,
            options= self.options,
        )
        res.x = (results['scale_homothety'] * res.x) + results['scale_translation']

        t_opt = perf_counter()-t0
        results["optimization_time"] = t_opt
        results["best_cost"] = res.fun
        results["best_position"]= res.x

        results.update(res)
        # hess_inv cannot be converted to json
        if "hess_inv" in results:
            del results["hess_inv"]
        results["status"] = "Completed"
        return results

class PSO_optimizer(Optimizer):
    def __init__(
        self,
        n_particles=5,
        dimensions=50,
        options=None,
        maxiter=1,
        n_processes=None,
        bounds=(0, 0),
        init_pos=None,
        print_time=False,
        opt_type="global",
        static=False,
        ftol=None,
        ftol_iter=1,
        bh_strategy="nearest",
        oh_strategy=None,
        save_results=False,
        saving_file="pso_results.json",
        comment=None,
    ):
        """
        Perform a Particle swarm optimization (PSO) on with a defined cost function using pyswarms
        library[1]

        Parameters
        ----------
        n_particles           	: int
            number of particle of the swarm, by default 5
        dimensions              : int
            number of dimensions of each particle
        options                 : dict
            hyperparameter of the PSO
        maxiter                    : int
            number of iteration of the PSO
        n_processes             : int
            number of process used to parallelize cost calculation, by default None
        bounds                  : tupple
            bounds of the particle, if equal no bounds, by default (0, 0)
        init_pos                : array
            initial position of the particles if None random, by default None
        print_time              : bool
            if True, print the optimisation time, by default True
        opt_type                : str
            Neightboorhood type, by default "global"
            type possibly:
                    "global"                : Global best PSO (star topology)
                    "local"                 : Local best PSO (ring topology)
        static      : bool
            if False and opt_type is local, update the neigthboorhood of each particle every iterations
        bh_strategy             : str
            out of bound position strategy for pyswarms optimizer [2]:
                    "nearest"               : Round the value to the nearest bound (default)
                    "periodic"              : set to the modulus of the value between the two bounds
                    "random"                : set to a random value
                    "shrink"                : reduce the velocity to finish land on the bound
                    "reflective"            : mirror the position form inside to ouside the bounds
                    "intermediate"          : set to intermediate value between previous pos and bound
        oh_strategy             : dict (like {"w":str, "c1":str, "c2":str})
            Dynamic options strategy for pyswarms optimizer [3], if None static options,
            by default None:
                    "exp_decay"             : Decreases the parameter exponentially between limits
                    "lin_variation"         : Decreases/increases the parameter linearly between limits
                    "nonlin_mod"            : Decreases/increases the parameter between limits
                                            according to a nonlinear modulation index
                    "rand"                  : takes a uniform random value between limits
        ftol                    : float
            relative error in objective_func(best_pos) acceptable for convergence, if None -np.inf
            default None
        ftol                    : int
            number of iterations over which the relative error in objective_func(best_pos) is
            acceptable for convergence, by default 1
        save_results            : bool
            save or not the output in a .json file, by default False
        saving_file             : str
            name of the file on wich the output should be saved, by default "pso_results.json"

        Returns
        -------
        results     : optim_results
            contains all the parameters and outputs of the PSO

        Note
        ----
        links to pyswarms doc:
        [1] https://pyswarms.readthedocs.io/en/latest/index.html
        [2] https://pyswarms.readthedocs.io/en/latest/api/pyswarms.handlers.html
        """
        super().__init__("PSO")
        self.swarm_optimizer = True
        self.n_particles = n_particles
        self.dimensions = dimensions
        self.options = options
        self.maxiter = maxiter
        self.n_processes = n_processes
        self.bounds = bounds
        self.init_pos = init_pos
        self.print_time = print_time
        self.opt_type = opt_type
        self.static = static
        self.ftol = ftol
        self.ftol_iter = ftol_iter
        self.bh_strategy = bh_strategy
        self.oh_strategy = oh_strategy
        self.save_results = save_results
        self.saving_file = saving_file
        self.comment = comment

    def __nrv2pyswarms_bounds(self):
        if np.size(self.bounds)==2*self.dimensions:
            if isinstance(self.bounds[0], tuple) and np.size(self.bounds[0])==2:
                max_bound = np.zeros(self.dimensions)
                min_bound = np.zeros(self.dimensions)
                for i, bd in enumerate(self.bounds):
                    max_bound[i] = max(bd)
                    min_bound[i] = min(bd)
                return (min_bound, max_bound)
        if np.size(self.bounds)>2:
            return self.bounds
        elif self.bounds[0] == self.bounds[1]:
            return None
        else:
            max_bound = max(self.bounds) * np.ones(self.dimensions)
            min_bound = min(self.bounds) * np.ones(self.dimensions)
            return (min_bound, max_bound)
    
    def minimize(self, f_swarm, **kwargs):
        """
        Perform a Particle swarm optimization

        Parameters
        ----------
        cost_function_swarm     : func
            function taking in parameter a swarm (dim-dimensions array) and returning the cost for each
            particle (1-dimensionnal array)
        kwargs                  : dict
            containing parameters to change to class (PSO_optimizer.__init__)
        """
        results = super().minimize(f_swarm, **kwargs)

        verbose = parameters.get_nrv_verbosity() > 2
        # initial position
        if not np.iterable(self.init_pos):
            self.init_pos = None

        # create pyswarms bounds
        psbounds = self.__nrv2pyswarms_bounds()

        if self.opt_type.lower()=="local":
            if not self.options or len(self.options) <5:
                self.options = {"c1": 0.5, "c2": 0.5, "w":0.5, "k" : 5, "p" : 2}
            topology = Ring(static=self.static)
        # Check number of pcu
        else:
            if not self.options:
                self.options = {"c1": 0.5, "c2": 0.5, "w":0.5}
            topology = Star()

        if self.n_processes != None and self.n_processes > cpu_count()-1:
            answer = input("Number of process higher than number of cpu\n"+
                "continue with one process (Y/n)\n")
            if answer == "Y":
                self.n_processes = None
            else:
                print("Terminated")
                return(False)

        # Initialize results directory
        if self.opt_type.lower()!="global":
            results["optimization_parameters"]["static"]=self.static

        if self.comment is not None: 
            results["comment"] = self.comment

        if self.save_results:
            with open(self.saving_file, "w") as outfile:
                json.dump(results, outfile)

        if self.ftol is None:
            ## As np.inf cannot be encode by json encoding ftol is used as a local varialble
            ftol = -np.inf
            self.ftol_iter = 1
        else:
            ftol = self.ftol

        try:
            # Call instance of PSO
            optimizer = ps.single.general_optimizer.GeneralOptimizerPSO(n_particles=self.n_particles,
                dimensions=self.dimensions, options=self.options, oh_strategy=self.oh_strategy, bounds=psbounds, ftol=ftol,
                ftol_iter=self.ftol_iter, init_pos=self.init_pos, bh_strategy=self.bh_strategy, topology=topology)
            optimizer.rep = rep_nrv
            optimizer.rep._load_defaults()
            # Perform optimization
            t0 = perf_counter()
            cost, pos = optimizer.optimize(f_swarm, iters=self.maxiter, n_processes=self.n_processes,
                verbose=verbose)
            t_opt = perf_counter()-t0

            pos_history = np.array(optimizer.pos_history)
            self.nit = np.shape(pos_history[:,0,:])[0]

            if self.print_time:
                print("the cumputing time for ", self.nit, " iterations and ", self.n_particles,
                    " particles is : ",t_opt,"s")

            # Save results
            results["nit"] = self.nit
            results["nfev"] = self.nit * self.n_particles
            results["status"] = "Completed"
            results["best_position"] = pos.tolist()
            results["x"] = results["best_position"]
            results["best_cost"] = cost

            results["optimization_time"] = t_opt
            results["cost_history"] = optimizer.cost_history
            results["neighbor_bestc"] = optimizer.mean_neighbor_history
            results["personal_bestc"] = optimizer.mean_pbest_history

            velocity_history = np.array(optimizer.velocity_history)
            pos_history = np.array(optimizer.pos_history)
            for i in range(self.n_particles):
                results["position"+str(i+1)] = pos_history[:,i,:].tolist()
                results["velocity"+str(i+1)] = velocity_history[:,i,:].tolist()
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)
        except KeyboardInterrupt:
            results["status"] = "Interrupted"
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)
            raise KeyboardInterrupt
            sys.exit(1)
        except:
            results["status"] = "Failed"
            results["Error_from_prompt"] = traceback.format_exc()
            print(results["Error_from_prompt"])
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)

        return results
